import os

import pandas as pd
import pytest

import audeer
import audformat.testing

import audb


DB_NAME = "test_load_on_demand"
DB_VERSION = "1.0.0"


@pytest.fixture(
    scope="module",
    autouse=True,
)
def dbs(tmpdir_factory, persistent_repository, storage_format):
    # Collect single database paths
    # and return them in the end
    paths = {}

    # publish 1.0.0

    db_root = tmpdir_factory.mktemp(DB_VERSION)
    paths[DB_VERSION] = db_root

    db = audformat.testing.create_db(minimal=True)
    db.name = DB_NAME
    db.schemes["scheme"] = audformat.Scheme()
    audformat.testing.add_table(
        db,
        "table1",
        "filewise",
        num_files=[0, 1, 2],
    )
    audformat.testing.add_table(
        db,
        "table2",
        "filewise",
        num_files=[1, 2, 3],
    )
    audformat.testing.add_misc_table(
        db,
        "misc-in-scheme",
        pd.Index([0, 1, 2], dtype="Int64", name="idx"),
        columns={"emotion": ("scheme", None)},
    )
    audformat.testing.add_misc_table(
        db,
        "misc-not-in-scheme",
        pd.Index([0, 1, 2], dtype="Int64", name="idx"),
        columns={"emotion": ("scheme", None)},
    )
    db.schemes["misc"] = audformat.Scheme(
        "int",
        labels="misc-in-scheme",
    )
    db.attachments["file"] = audformat.Attachment("file.txt")
    db.attachments["folder"] = audformat.Attachment("folder/")
    audeer.mkdir(db_root, "folder")
    audeer.touch(db_root, "file.txt")
    audeer.touch(db_root, "folder/file1.txt")
    audeer.touch(db_root, "folder/file2.txt")
    db.save(db_root, storage_format=storage_format)
    audformat.testing.create_audio_files(db)

    audb.publish(
        db_root,
        DB_VERSION,
        persistent_repository,
        verbose=False,
    )

    return paths


def test_load_only_metadata(dbs, storage_format):
    db_original = audformat.Database.load(dbs[DB_VERSION])

    db = audb.load(
        DB_NAME,
        version=DB_VERSION,
        only_metadata=True,
        full_path=False,
        verbose=False,
    )

    assert db["table1"] == db_original["table1"]
    assert db["table2"] == db_original["table2"]
    assert db["misc-in-scheme"] == db_original["misc-in-scheme"]
    assert db["misc-not-in-scheme"] == db_original["misc-not-in-scheme"]
    pd.testing.assert_index_equal(db.files, db_original.files)
    for attachment in db.attachments:
        path = audeer.path(db.root, db.attachments[attachment].path)
        assert not os.path.exists(path)
    for file in list(db.files):
        path = audeer.path(db.root, file)
        assert not os.path.exists(path)
    assert not db.meta["audb"]["complete"]

    # Delete table1
    # to force downloading from backend again
    os.remove(os.path.join(db.meta["audb"]["root"], f"db.table1.{storage_format}"))
    os.remove(os.path.join(db.meta["audb"]["root"], "db.table1.pkl"))
    db = audb.load(
        DB_NAME,
        version=DB_VERSION,
        only_metadata=True,
        full_path=False,
        verbose=False,
    )
    assert db["table1"] == db_original["table1"]
    assert db["table2"] == db_original["table2"]
    assert db["misc-in-scheme"] == db_original["misc-in-scheme"]
    assert db["misc-not-in-scheme"] == db_original["misc-not-in-scheme"]
    pd.testing.assert_index_equal(db.files, db_original.files)
    assert not db.meta["audb"]["complete"]

    # Load whole database
    db = audb.load(
        DB_NAME,
        version=DB_VERSION,
        full_path=False,
        verbose=False,
    )
    db_original.meta = []
    db.meta = []
    assert db == db_original


@pytest.mark.parametrize(
    "attachments, "
    "media, "
    "tables, "
    "expected_attachments, "
    "expected_media, "
    "expected_tables, "
    "complete",
    [
        (
            None,
            None,
            None,
            ["file", "folder"],
            [
                "audio/000.wav",
                "audio/001.wav",
                "audio/002.wav",
                "audio/003.wav",
            ],
            ["misc-in-scheme", "misc-not-in-scheme", "table1", "table2"],
            True,
        ),
        (
            [],
            None,
            None,
            [],
            [
                "audio/000.wav",
                "audio/001.wav",
                "audio/002.wav",
                "audio/003.wav",
            ],
            ["misc-in-scheme", "misc-not-in-scheme", "table1", "table2"],
            False,
        ),
        (
            None,
            [],
            None,
            ["file", "folder"],
            [],
            ["misc-in-scheme", "misc-not-in-scheme", "table1", "table2"],
            False,
        ),
        (
            None,
            None,
            [],
            ["file", "folder"],
            [],
            ["misc-in-scheme"],  # misc table used in scheme is still loaded
            False,
        ),
        (
            [],
            [],
            [],
            [],
            [],
            ["misc-in-scheme"],
            False,
        ),
        (
            [],
            [],
            None,
            [],
            [],
            ["misc-in-scheme", "misc-not-in-scheme", "table1", "table2"],
            False,
        ),
        (
            [],
            None,
            [],
            [],
            [],
            ["misc-in-scheme"],
            False,
        ),
        (
            None,
            [],
            [],
            ["file", "folder"],
            [],
            ["misc-in-scheme"],
            False,
        ),
        (
            "folder",
            None,
            None,
            ["folder"],
            [
                "audio/000.wav",
                "audio/001.wav",
                "audio/002.wav",
                "audio/003.wav",
            ],
            ["misc-in-scheme", "misc-not-in-scheme", "table1", "table2"],
            False,
        ),
        (
            ["file"],
            None,
            ["table1"],
            ["file"],
            [
                "audio/000.wav",
                "audio/001.wav",
                "audio/002.wav",
            ],
            ["misc-in-scheme", "table1"],
            False,
        ),
        (
            ["file", "folder"],
            None,
            ".*1",
            ["file", "folder"],
            [
                "audio/000.wav",
                "audio/001.wav",
                "audio/002.wav",
            ],
            ["misc-in-scheme", "table1"],
            False,
        ),
        (
            None,
            ".3.wav",
            None,
            ["file", "folder"],
            [
                "audio/003.wav",
            ],
            ["misc-in-scheme", "misc-not-in-scheme", "table1", "table2"],
            False,
        ),
    ],
)
def test_load_filter(
    tmpdir,
    attachments,
    media,
    tables,
    expected_attachments,
    expected_media,
    expected_tables,
    complete,
):
    db = audb.load(
        DB_NAME,
        version=DB_VERSION,
        attachments=attachments,
        media=media,
        tables=tables,
        full_path=False,
        cache_root=tmpdir,
        verbose=False,
    )
    assert list(db) == expected_tables
    assert list(db.files) == expected_media
    assert list(db.attachments) == ["file", "folder"]
    for attachment in db.attachments:
        path = audeer.path(db.root, db.attachments[attachment].path)
        if attachment in expected_attachments:
            assert os.path.exists(path)
            for file in db.attachments[attachment].files:
                assert os.path.exists(audeer.path(db.root, file))
        else:
            assert not os.path.exists(path)
    assert db.meta["audb"]["complete"] == complete



import random
import time

import pandas as pd
import tabulate

import audeer

import audb


random.seed(1)

cache = audeer.mkdir("./cache")

CACHE_EXT: str = "pkl"
CACHE_EXT: str = "parquet"
PARQUET_SAVE_OPTS: dict = {"engine": "pyarrow"}
# dtypes : list = ["string", "object", "pyarrow"]
dtypes: list = [
    "polars",
]


def set_dependency_module():
    r"""Monkeypatch dependency modult to use polars module."""
    import polars as pl

    from audb.core import define

    depend_index_colname = "file"
    depend_index_dtype = pl.datatypes.Object
    depend_field_dtypes = dict(
        zip(
            define.DEPEND_FIELD_DTYPES.keys(),
            [
                pl.datatypes.String,
                pl.datatypes.Int32,
                pl.datatypes.Int32,
                pl.datatypes.String,
                pl.datatypes.Float64,
                pl.datatypes.String,
                pl.datatypes.Int32,
                pl.datatypes.Int32,
                pl.datatypes.Int32,
                pl.datatypes.String,
            ],
        )
    )

    audb.core.define.DEPEND_INDEX_COLNAME = depend_index_colname
    audb.core.define.DEPEND_FIELD_DTYPES_PANDAS = audb.core.define.DEPEND_FIELD_DTYPES
    audb.core.define.DEPEND_FIELD_DTYPES = depend_field_dtypes
    audb.core.define.DEPEND_INDEX_DTYPE_PANDAS = audb.core.define.DEPEND_INDEX_DTYPE
    audb.core.define.DEPEND_INDEX_DTYPE = depend_index_dtype

    import dependencies_polars

    audb.Dependencies = dependencies_polars.Dependencies


# === Dependencies load via pickle before monkey_patching ===
data_cache = audeer.path(cache, "df.pkl")
deps = audb.Dependencies()
deps.load(data_cache)

# save cache in parquet format as the polars load method depends on it
parquet_cache = audeer.path(cache, "df.parquet")
deps.save(parquet_cache)

file = "file-10.wav"
n_files = 10000
results = pd.DataFrame(columns=["polars"])
results.index.name = "method"
set_dependency_module()
dtype = "polars"

for dtype in dtypes:
    # load them
    deps = audb.Dependencies()
    deps.load(parquet_cache)
    _files = deps._df["file"][:n_files].to_list()

    # only string meanningful
    # expected_dtype = pl.String

    # assert deps._df["archive"].dtype == expected_dtype

    method = "Dependencies.__call__()"
    t0 = time.time()
    # deps()
    t = time.time() - t0
    results.at[method, dtype] = t

    # Access the index one time.
    # Further calls will be faster
    file in deps

    method = f"Dependencies.__contains__({n_files} files)"
    t0 = time.time()
    [file in deps for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.__get_item__({n_files} files)"
    t0 = time.time()
    [deps[file] for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.__len__()"
    t0 = time.time()
    len(deps)
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.__str__()"
    t0 = time.time()
    str(deps)
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.archives"
    t0 = time.time()
    deps.archives
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.attachments"
    t0 = time.time()
    deps.attachments
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.attachment_ids"
    t0 = time.time()
    deps.attachment_ids
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.files"
    t0 = time.time()
    deps.files
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.media"
    t0 = time.time()
    deps.media
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.removed_media"
    t0 = time.time()
    deps.removed_media
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.table_ids"
    t0 = time.time()
    deps.table_ids
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.tables"
    t0 = time.time()
    deps.tables
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.archive({n_files} files)"
    t0 = time.time()
    [deps.archive(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.bit_depth({n_files} files)"
    t0 = time.time()
    [deps.bit_depth(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.channels({n_files} files)"
    t0 = time.time()
    [deps.channels(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.checksum({n_files} files)"
    t0 = time.time()
    [deps.checksum(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.duration({n_files} files)"
    t0 = time.time()
    [deps.duration(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.format({n_files} files)"
    t0 = time.time()
    [deps.format(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.removed({n_files} files)"
    t0 = time.time()
    [deps.removed(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.sampling_rate({n_files} files)"
    t0 = time.time()
    [deps.sampling_rate(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.type({n_files} files)"
    t0 = time.time()
    [deps.type(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.version({n_files} files)"
    t0 = time.time()
    [deps.version(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    # -------------------------------------------------------------------------

    # TODO: Reimplement
    method = "Dependencies._add_attachment()"
    t0 = time.time()
    deps._add_attachment("attachment.txt", "1.0.0", "archive", "checksum")
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies._add_media({n_files} files)"
    values = [
        (
            f"file-new-{n}.wav",  # file
            f"archive-new-{n}",  # archive
            16,  # bit_depth
            1,  # channels
            f"checksum-{n}",  # checksum
            0.4,  # duration
            "wav",  # format
            0,  # removed
            16000,  # sampling_rate
            1,  # type
            "1.0.0",  # version
        )
        for n in range(n_files)
    ]
    t0 = time.time()
    deps._add_media(values)
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies._add_meta()"
    t0 = time.time()
    deps._add_meta("db.new-table.csv", "1.0.0", "archive", "checksum")
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies._drop()"
    t0 = time.time()
    deps._drop(["file-90000.wav"])
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies._remove()"
    t0 = time.time()
    deps._remove(file)
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies._update_media()"
    t0 = time.time()
    deps._update_media(values)
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies._update_media_version({n_files} files)"
    t0 = time.time()
    deps._update_media_version([f"file-{n}.wav" for n in range(n_files)], "version")
    t = time.time() - t0
    results.at[method, dtype] = t

# ===== Print results =====
table = tabulate.tabulate(results, headers="keys", tablefmt="github", floatfmt=".3f")
fp_results = audeer.path(cache, "results_polars.csv")
results.to_csv(fp_results)

print(table)



from __future__ import annotations

from collections.abc import Sequence
import os
import shutil
import warnings

import numpy as np

import audeer
import audiofile
import audobject
import audresample

from audb.core import define


class Flavor(audobject.Object):
    r"""Database flavor.

    Helper class used by :meth:`audb.load`
    to convert media files to the desired format.
    It stores the meta information about a flavor
    and offers a convenient way to convert files to it.

    As the following example shows,
    it can also be used to convert files
    that are not part of a database:

    .. code-block::  python

        original_file = "/org/path/file.flac"
        converted_file = "/new/path/file.wav"

        # convert file to 16 kHz
        flavor = Flavor(sampling_rate=16000)
        flavor(original_file, converted_file)

    Args:
        bit_depth: sample precision, one of ``16``, ``24``, ``32``
        channels: channel selection, see :func:`audresample.remix`
        format: file format, one of ``'flac'``, ``'wav'``
        mixdown: apply mono mix-down on selection
        sampling_rate: sampling rate in Hz, one of
            ``8000``, ``16000``, ``22050``, ``24000``, ``44100``, ``48000``

    Raises:
        ValueError: if a non-supported ``bit_depth``,
            ``format``,
            or ``sampling_rate``
            is requested

    """

    def __init__(
        self,
        *,
        bit_depth: int = None,
        channels: int | Sequence[int] = None,
        format: str = None,
        mixdown: bool = False,
        sampling_rate: int = None,
    ):
        if bit_depth is not None:
            bit_depth = int(bit_depth)
            if bit_depth not in define.BIT_DEPTHS:
                raise ValueError(
                    f"Bit depth has to be one of "
                    f"{define.BIT_DEPTHS}, not {bit_depth}."
                )

        if channels is not None:
            channels = audeer.to_list(channels)
            if len(channels) < 2:
                mixdown = False

        if format is not None:
            if format not in define.FORMATS:
                raise ValueError(
                    f"Format has to be one of " f"{define.FORMATS}, not '{format}'."
                )

        if sampling_rate is not None:
            sampling_rate = int(sampling_rate)
            if sampling_rate == 22500:
                message = (
                    "A sampling rate of 22500 Hz is deprecated "
                    "and will be removed with version 1.13.0 of audb. "
                    "Use 22050 Hz instead."
                )
                warnings.warn(message, category=UserWarning, stacklevel=2)
            elif sampling_rate not in define.SAMPLING_RATES:
                raise ValueError(
                    f"Sampling_rate has to be one of "
                    f"{define.SAMPLING_RATES}, not {sampling_rate}."
                )

        self.bit_depth = bit_depth
        r"""Sample precision."""
        self.channels = channels
        r"""Selected channels."""
        self.format = format
        r"""File format."""
        self.mixdown = mixdown
        r"""Apply mixdown."""
        self.sampling_rate = sampling_rate
        r"""Sampling rate in Hz."""

    def destination(
        self,
        file: str,
    ) -> str:
        r"""Return converted file path.

        The file path will only change if the file is converted to a different
        format.

        Args:
            file: path to input file

        Returns:
            path to output file

        """
        if self.format is not None:
            format = audeer.file_extension(file).lower()
            if format != self.format:
                file = f"{file[:-len(format)]}{self.format}"
        return file

    def path(
        self,
        name: str,
        version: str,
    ) -> str:
        r"""Flavor path.

        Args:
            name: database name
            version: version string

        Returns:
            relative path

        """
        return os.path.join(name, version, self.short_id)

    @property
    def short_id(
        self,
    ) -> str:
        r"""Short flavor ID.

        This just truncates the ID
        to its last eight characters.

        """
        return self.id[-8:]

    def _check_convert(
        self,
        file: str,
        bit_depth: int | None,
        channels: int | None,
        sampling_rate: int | None,
    ) -> bool:
        r"""Check if file needs to be converted to flavor."""
        format = audeer.file_extension(file).lower()

        # format change
        if self.format is not None:
            if self.format != format:
                return True

        convert = False

        # precision change
        if not convert and self.bit_depth is not None:
            bit_depth = bit_depth or audiofile.bit_depth(file)
            if self.bit_depth != bit_depth:
                convert = True

        # mixdown and channel selection
        if not convert and self.mixdown or self.channels is not None:
            channels = channels or audiofile.channels(file)
            if self.mixdown and channels != 1:
                convert = True
            elif list(range(channels)) != self.channels:
                convert = True

        # sampling rate change
        if not convert and self.sampling_rate is not None:
            sampling_rate = sampling_rate or audiofile.sampling_rate(file)
            if self.sampling_rate != sampling_rate:
                convert = True

        if convert and format not in define.FORMATS:
            raise RuntimeError(
                f"You have to specify the 'format' argument "
                f"to convert '{file}' "
                f"to the specified flavor "
                f"as we cannot write to {format.upper()} files."
            )

        return convert

    def _remix(
        self,
        signal: np.ndarray,
    ) -> np.ndarray:
        if self.channels is None and not self.mixdown:
            return signal
        else:
            return audresample.remix(
                signal,
                self.channels,
                self.mixdown,
                upmix="repeat",
            )

    def _resample(
        self,
        signal: np.ndarray,
        sampling_rate: int,
    ) -> (np.ndarray, int):
        r"""Resample signal to flavor."""
        if (self.sampling_rate is not None) and (sampling_rate != self.sampling_rate):
            signal = audresample.resample(
                signal,
                sampling_rate,
                self.sampling_rate,
            )
            sampling_rate = self.sampling_rate
        return signal, sampling_rate

    def __call__(
        self,
        src_path: str,
        dst_path: str,
        *,
        src_bit_depth: int = None,
        src_channels: int = None,
        src_sampling_rate: int = None,
    ):
        r"""Convert file to flavor.

        If ``bit_depth``, ``channels`` or ``sampling_rate``
        of source signal are known, they can be provided.
        Otherwise, they will be computed using :mod:`audiofile`.

        Args:
            src_path: path to input file
            dst_path: path to output file
            src_bit_depth: bit depth
            src_channels: number of channels
            src_sampling_rate: sampling rate in Hz

        Raises:
            ValueError: if extension of output file does not match the
                format of the flavor
            RuntimeError: if a conversion is requested,
                but no output format is specified,
                and the input format is not WAV or FLAC

        """
        src_path = audeer.path(src_path)
        dst_path = audeer.path(dst_path)

        # verify that extension matches the output format
        src_ext = audeer.file_extension(src_path).lower()
        dst_ext = audeer.file_extension(dst_path).lower()
        expected_ext = self.format or src_ext
        if expected_ext != dst_ext:
            raise ValueError(
                f"Extension of output file is '{dst_ext}', "
                f"but should be '{expected_ext}' "
                "to match the format of the converted file."
            )

        if not self._check_convert(
            src_path, src_bit_depth, src_channels, src_sampling_rate
        ):
            # file already satisfies flavor
            if src_path != dst_path:
                shutil.copy(src_path, dst_path)

        else:
            # convert file to flavor
            signal, sampling_rate = audiofile.read(src_path, always_2d=True)
            signal = self._remix(signal)
            signal, sampling_rate = self._resample(signal, sampling_rate)
            bit_depth = (
                self.bit_depth or src_bit_depth or audiofile.bit_depth(src_path) or 16
            )
            audiofile.write(
                dst_path,
                signal,
                sampling_rate,
                bit_depth=bit_depth,
            )



import hashlib
import os
import pickle
import random
import string
import time

import memray
import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.csv as csv
import pyarrow.parquet as parquet
import tabulate

import audeer

import audb


random.seed(1)

cache = audeer.mkdir("./cache")
result_dir = audeer.path("./results")
audeer.rmdir(result_dir)
audeer.mkdir(result_dir)


def astype(df, dtype):
    r"""Convert dataframe to desired format."""
    if dtype == "object":
        # Use `object` to represent strings
        df["archive"] = df["archive"].astype("object")
        df["bit_depth"] = df["bit_depth"].astype("int32")
        df["channels"] = df["channels"].astype("int32")
        df["checksum"] = df["checksum"].astype("object")
        df["duration"] = df["duration"].astype("float64")
        df["format"] = df["format"].astype("object")
        df["removed"] = df["removed"].astype("int32")
        df["sampling_rate"] = df["sampling_rate"].astype("int32")
        df["type"] = df["type"].astype("int32")
        df["version"] = df["version"].astype("object")
        df.index = df.index.astype(audb.core.define.DEPENDENCY_INDEX_DTYPE)
    elif dtype == "string":
        # Use `string` to represent strings
        df["archive"] = df["archive"].astype("string")
        df["bit_depth"] = df["bit_depth"].astype("int32")
        df["channels"] = df["channels"].astype("int32")
        df["checksum"] = df["checksum"].astype("string")
        df["duration"] = df["duration"].astype("float64")
        df["format"] = df["format"].astype("string")
        df["removed"] = df["removed"].astype("int32")
        df["sampling_rate"] = df["sampling_rate"].astype("int32")
        df["type"] = df["type"].astype("int32")
        df["version"] = df["version"].astype("string")
        df.index = df.index.astype(audb.core.define.DEPENDENCY_INDEX_DTYPE)
    elif dtype == "pyarrow":
        # Use `pyarrow` to represent all dtypes
        df["archive"] = df["archive"].astype("string[pyarrow]")
        df["bit_depth"] = df["bit_depth"].astype("int32[pyarrow]")
        df["channels"] = df["channels"].astype("int32[pyarrow]")
        df["checksum"] = df["checksum"].astype("string[pyarrow]")
        df["duration"] = df["duration"].astype("float64[pyarrow]")
        df["format"] = df["format"].astype("string[pyarrow]")
        df["removed"] = df["removed"].astype("int32[pyarrow]")
        df["sampling_rate"] = df["sampling_rate"].astype("int32[pyarrow]")
        df["type"] = df["type"].astype("int32[pyarrow]")
        df["version"] = df["version"].astype("string[pyarrow]")
        df.index = df.index.astype(audb.core.define.DEPENDENCY_INDEX_DTYPE)
    return df


# Create dependency dataframe
data_cache = audeer.path(cache, "df.pkl")
num_rows = 1000000
if not os.path.exists(data_cache):
    bit_depths = [0, 16, 24]
    channels = [0, 1, 2]
    formats = ["csv", "wav", "txt"]
    sampling_rates = [0, 16000, 44100]
    types = [0, 1, 2]
    versions = ["1.0.0", "1.1.0"]
    records = [
        {
            "file": f"file-{n}.wav",
            "archive": f"archive-{n}",
            "bit_depth": random.choices(bit_depths, weights=[0.1, 0.8, 0.1])[0],
            "channels": random.choices(channels, weights=[0.1, 0.8, 0.1])[0],
            "checksum": hashlib.md5(
                pickle.dumps(random.choice(string.ascii_letters))
            ).hexdigest(),
            "duration": 10 * random.random(),
            "format": random.choices(formats, weights=[0.1, 0.8, 0.1])[0],
            "removed": random.choices([0, 1], weights=[0.1, 0.9])[0],
            "sampling_rate": random.choices(sampling_rates, weights=[0.1, 0.8, 0.1])[0],
            "type": random.choices(types, weights=[0.1, 0.8, 0.1])[0],
            "version": random.choices(versions, weights=[0.2, 0.8])[0],
        }
        for n in range(num_rows)
    ]
    df = pd.DataFrame.from_records(records)
    df = df.astype(audb.core.define.DEPENDENCY_TABLE)
    df.set_index("file", inplace=True)
    df.index.name = None
    df.index = df.index.astype(audb.core.define.DEPENDENCY_INDEX_DTYPE)
    df.to_pickle(data_cache)
else:
    df = pd.read_pickle(data_cache)

# ===== Data type mappings =====
dtype_mapping = {
    # pandas string as `string`
    "string": {
        "archive": "string",
        "bit_depth": "int32",
        "channels": "int32",
        "checksum": "string",
        "duration": "float64",
        "format": "string",
        "removed": "int32",
        "sampling_rate": "int32",
        "type": "int32",
        "version": "string",
    },
    # pandas string as `object`
    "object": {
        "archive": "object",
        "bit_depth": "int32",
        "channels": "int32",
        "checksum": "object",
        "duration": "float64",
        "format": "object",
        "removed": "int32",
        "sampling_rate": "int32",
        "type": "int32",
        "version": "object",
    },
    # pandas pyarrow dtypes
    "pyarrow": {
        "archive": "string[pyarrow]",
        "bit_depth": "int32[pyarrow]",
        "channels": "int32[pyarrow]",
        "checksum": "string[pyarrow]",
        "duration": "float64[pyarrow]",
        "format": "string[pyarrow]",
        "removed": "int32[pyarrow]",
        "sampling_rate": "int32[pyarrow]",
        "type": "int32[pyarrow]",
        "version": "string[pyarrow]",
    },
}
# pyarrow dtype schema
pyarrow_schema = pa.schema(
    [
        ("file", pa.string()),
        ("archive", pa.string()),
        ("bit_depth", pa.int32()),
        ("channels", pa.int32()),
        ("checksum", pa.string()),
        ("duration", pa.float64()),
        ("format", pa.string()),
        ("removed", pa.int32()),
        ("sampling_rate", pa.int32()),
        ("type", pa.int32()),
        ("version", pa.string()),
    ]
)

# ===== Benchmark storing data =====
folder = audeer.path(cache, "files")
audeer.rmdir(folder)
audeer.mkdir(folder)

dtypes = ["object", "string", "pyarrow"]

# ===== WRITING ===========================================================
results = pd.DataFrame(columns=["csv", "pickle", "parquet"])
results.index.name = "method"

print()
print("* Write to CSV file.", end="", flush=True)
file = audeer.path(folder, "df.csv")

for dtype in dtypes:
    _df = df.copy()
    _df = astype(_df, dtype)
    t0 = time.time()
    df.to_csv(file)
    t = time.time() - t0
    method = f"pd.DataFrame[{dtype}]"
    results.at[method, "csv"] = t
    print(".", end="", flush=True)

# Use object and copy to pyarrow before writing
_df = df.copy()
_df = astype(_df, "object")
t0 = time.time()
_df = astype(_df, "pyarrow")
_df.to_csv(file)
t = time.time() - t0
method = "pd.DataFrame[object] -> pd.DataFrame[pyarrow]"
results.at[method, "csv"] = t
print(".", end="", flush=True)

for dtype in dtypes:
    _df = df.copy()
    _df = astype(_df, dtype)
    t0 = time.time()
    _df.index = _df.index.rename("file")
    _df = _df.reset_index()
    table = pa.Table.from_pandas(_df, preserve_index=False, schema=pyarrow_schema)
    _columns = table.column_names
    _columns = ["" if c == "file" else c for c in _columns]
    _table = table.rename_columns(_columns)
    csv.write_csv(
        _table,
        file,
        write_options=csv.WriteOptions(quoting_style="none"),
    )
    t = time.time() - t0
    method = f"pd.DataFrame[{dtype}] -> pa.Table"
    results.at[method, "csv"] = t
    print(".", end="", flush=True)

# Use object and copy to pyarrow before writing
_df = df.copy()
_df = astype(_df, "object")
t0 = time.time()
_df = astype(_df, "pyarrow")
_df.index = _df.index.rename("file")
_df = _df.reset_index()
table = pa.Table.from_pandas(_df, preserve_index=False, schema=pyarrow_schema)
_columns = table.column_names
_columns = ["" if c == "file" else c for c in _columns]
_table = table.rename_columns(_columns)
csv.write_csv(
    _table,
    file,
    write_options=csv.WriteOptions(quoting_style="none"),
)
t = time.time() - t0
method = "pd.DataFrame[object] -> pd.DataFrame[pyarrow] -> pa.Table"
results.at[method, "csv"] = t
print(".", end="", flush=True)

t0 = time.time()
_table = table.rename_columns(_columns)
csv.write_csv(
    _table,
    file,
    write_options=csv.WriteOptions(quoting_style="none"),
)
t = time.time() - t0
method = "pa.Table"
results.at[method, "csv"] = t
print(".")

# -------------------------------------------------------------------------
print("* Write to PKL file.", end="", flush=True)

for dtype in dtypes:
    file = audeer.path(folder, f"df-{dtype}.pkl")
    _df = df.copy()
    _df = astype(_df, dtype)
    t0 = time.time()
    _df.to_pickle(file, protocol=4)
    t = time.time() - t0
    method = f"pd.DataFrame[{dtype}]"
    results.at[method, "pickle"] = t
    print(".", end="", flush=True)

# Use object and copy to pyarrow before writing
_df = df.copy()
_df = astype(_df, "object")
file = audeer.path(folder, "df-pyarrow.pkl")
t0 = time.time()
_df = astype(_df, "pyarrow")
_df.to_pickle(file, protocol=4)
t = time.time() - t0
method = "pd.DataFrame[object] -> pd.DataFrame[pyarrow]"
results.at[method, "pickle"] = t
print(".")

# -------------------------------------------------------------------------
print("* Write to PARQUET file.", end="", flush=True)
file = audeer.path(folder, "df.parquet")

for dtype in dtypes:
    _df = df.copy()
    _df = astype(_df, dtype)
    t0 = time.time()
    _df.to_parquet(file, engine="pyarrow")
    t = time.time() - t0
    method = f"pd.DataFrame[{dtype}]"
    results.at[method, "parquet"] = t
    print(".", end="", flush=True)

t0 = time.time()
parquet.write_table(table, file)
t = time.time() - t0
method = "pa.Table"
results.at[method, "parquet"] = t
print(".")
print()

# ===== Print results =====
print(f"Results for writing {num_rows} lines.")
print()
results = results.replace(np.nan, None)
results = tabulate.tabulate(results, headers="keys", tablefmt="github", floatfmt=".3f")
print(results)


# ===== READING ===========================================================
results = pd.DataFrame(columns=["csv", "pickle", "parquet"])
results.index.name = "method"


def method_to_filename(method):
    return method.replace(" ", "-").replace("[", "-").replace("]", "-").replace(">", "")


print()
print("* Read CSV file.", end="", flush=True)
file = audeer.path(folder, "df.csv")

for engine in [None, "c", "pyarrow"]:
    if engine is None:
        arrow = "---->"
    elif engine == "pyarrow":
        arrow = "-pa->"
    elif engine == "c":
        arrow = "-c-->"

    for dtype in dtypes:
        method = f"{arrow} pd.DataFrame[{dtype}]"
        memray_file = f"memray-csv-{method_to_filename(method)}.bin"
        with memray.Tracker(audeer.path(result_dir, memray_file)):
            t0 = time.time()
            index_col = 0
            _df = pd.read_csv(
                file,
                index_col=index_col,
                na_filter=False,
                dtype=dtype_mapping[dtype],
                header=0,
                engine=engine,
            )
            t = time.time() - t0
        results.at[method, "csv"] = t
        if dtype == "pyarrow":
            dtype = "string[pyarrow]"
        _df.index = _df.index.astype(audb.core.define.DEPENDENCY_INDEX_DTYPE)
        assert _df.archive.dtype == dtype
        print(".", end="", flush=True)

    # Convert pyarrow dtypes to object
    method = f"{arrow} pd.DataFrame[pyarrow] -> pd.DataFrame[object]"
    memray_file = f"memray-csv-{method_to_filename(method)}.bin"
    with memray.Tracker(audeer.path(result_dir, memray_file)):
        t0 = time.time()
        index_col = 0
        _df = pd.read_csv(
            file,
            index_col=index_col,
            na_filter=False,
            dtype=dtype_mapping["pyarrow"],
            header=0,
            engine=engine,
        )
        _df = astype(_df, "object")
        t = time.time() - t0
    results.at[method, "csv"] = t
    print(".", end="", flush=True)


for dtype in dtypes:
    method = f"----> pa.Table -> pd.DataFrame[{dtype}]"
    memray_file = f"memray-csv-{method_to_filename(method)}.bin"
    with memray.Tracker(audeer.path(result_dir, memray_file)):
        t0 = time.time()
        _table = csv.read_csv(
            file,
            read_options=csv.ReadOptions(
                column_names=table.column_names,
                skip_rows=1,
            ),
            convert_options=csv.ConvertOptions(column_types=pyarrow_schema),
        )
        if dtype == "object":
            types_mapper = None
        elif dtype == "string":
            types_mapper = {pa.string(): pd.StringDtype()}.get
        elif dtype == "pyarrow":
            # Ensure we use pd.StringDtype("pyarrow")
            # and not pd.ArrowDtype(pa.string())
            # see https://pandas.pydata.org/docs/user_guide/pyarrow.html
            types_mapper = {
                pa.string(): pd.StringDtype("pyarrow"),
                pa.int32(): pd.ArrowDtype(pa.int32()),
                pa.float64(): pd.ArrowDtype(pa.float64()),
            }.get
        _df = _table.to_pandas(
            # Speed up conversion,
            # increase memory usage by ~20 MB
            deduplicate_objects=False,
            types_mapper=types_mapper,
        )
        _df.set_index("file", inplace=True)
        _df.index.name = None
        t = time.time() - t0
    results.at[method, "csv"] = t
    if dtype == "pyarrow":
        assert _df.archive.dtype == "string[pyarrow]"
        assert _df.type.dtype == "int32[pyarrow]"
        assert _df.duration.dtype == "float64[pyarrow]"
    else:
        assert _df.archive.dtype == dtype
        assert _df.type.dtype == "int32"
        assert _df.duration.dtype == "float"
    print(".", end="", flush=True)

# Convert pyarrow dtypes to object
method = "----> pa.Table -> pd.DataFrame[pyarrow] -> pd.DataFrame[object]"
memray_file = f"memray-csv-{method_to_filename(method)}.bin"
with memray.Tracker(audeer.path(result_dir, memray_file)):
    t0 = time.time()
    _table = csv.read_csv(
        file,
        read_options=csv.ReadOptions(
            column_names=table.column_names,
            skip_rows=1,
        ),
        convert_options=csv.ConvertOptions(column_types=pyarrow_schema),
    )
    _df = _table.to_pandas(
        # Speed up conversion,
        # increase memory usage by ~20 MB
        deduplicate_objects=False,
        types_mapper=pd.ArrowDtype,
    )
    _df.set_index("file", inplace=True)
    _df.index.name = None
    _df = astype(_df, "object")
    t = time.time() - t0
results.at[method, "csv"] = t
print(".", end="", flush=True)

method = "----> pa.Table"
memray_file = f"memray-csv-{method_to_filename(method)}.bin"
with memray.Tracker(audeer.path(result_dir, memray_file)):
    t0 = time.time()
    _table = csv.read_csv(
        file,
        read_options=csv.ReadOptions(
            column_names=table.column_names,
            skip_rows=1,
        ),
        convert_options=csv.ConvertOptions(column_types=pyarrow_schema),
    )
    t = time.time() - t0
results.at[method, "csv"] = t
print(".")

# -------------------------------------------------------------------------
print("* Read PKL file.", end="", flush=True)

for dtype in dtypes:
    file = audeer.path(folder, f"df-{dtype}.pkl")
    method = f"----> pd.DataFrame[{dtype}]"
    memray_file = f"memray-pkl-{method_to_filename(method)}.bin"
    with memray.Tracker(audeer.path(result_dir, memray_file)):
        t0 = time.time()
        _df = pd.read_pickle(file)
        t = time.time() - t0
    results.at[method, "pickle"] = t
    print(".", end="", flush=True)

# Convert pyarrow dtypes to object
file = audeer.path(folder, "df-pyarrow.pkl")
method = "----> pd.DataFrame[pyarrow] -> pd.DataFrame[object]"
memray_file = f"memray-pkl-{method_to_filename(method)}.bin"
with memray.Tracker(audeer.path(result_dir, memray_file)):
    t0 = time.time()
    _df = pd.read_pickle(file)
    _df = astype(_df, "object")
    t = time.time() - t0
results.at[method, "pickle"] = t
print(".")

# -------------------------------------------------------------------------
print("* Read PARQUET file.", end="", flush=True)
file = audeer.path(folder, "df.parquet")

for dtype in dtypes:
    _df = df.copy()
    _df = astype(_df, dtype)
    _df.index.rename("file", inplace=True)
    _df = _df.reset_index()
    _df.to_parquet(file, index=False, engine="pyarrow")
    method = f"----> pd.DataFrame[{dtype}]"
    memray_file = f"memray-parquet-{method_to_filename(method)}.bin"
    with memray.Tracker(audeer.path(result_dir, memray_file)):
        t0 = time.time()
        _df = pd.read_parquet(file, engine="pyarrow")
        t = time.time() - t0
    results.at[method, "parquet"] = t
    print(".", end="", flush=True)

for dtype in dtypes:
    _df = df.copy()
    _df = astype(_df, dtype)
    _df.index.rename("file", inplace=True)
    _df = _df.reset_index()
    _df.to_parquet(file, index=False, engine="pyarrow")
    if dtype == "object":
        types_mapper = None
    elif dtype == "string":
        types_mapper = {pa.string(): pd.StringDtype()}.get
    elif dtype == "pyarrow":
        types_mapper = pd.ArrowDtype
    method = f"----> pa.Table -> pd.DataFrame[{dtype}]"
    memray_file = f"memray-parquet-{method_to_filename(method)}.bin"
    with memray.Tracker(audeer.path(result_dir, memray_file)):
        t0 = time.time()
        _table = parquet.read_table(file)
        _df = _table.to_pandas(
            # Speed up conversion,
            # increase memory usage by ~20 MB
            deduplicate_objects=False,
            types_mapper=types_mapper,
        )
        _df.set_index("file", inplace=True)
        _df.index.name = None
        t = time.time() - t0
    results.at[method, "parquet"] = t
    print(".", end="", flush=True)

results.at[method, "parquet"] = t
method = "----> pa.Table"
memray_file = f"memray-parquet-{method_to_filename(method)}.bin"
with memray.Tracker(audeer.path(result_dir, memray_file)):
    t0 = time.time()
    _table = parquet.read_table(file)
    t = time.time() - t0
    results.at[method, "parquet"] = t
print(".")
print()

# ===== Print results =====
print(f"Results for reading {num_rows} lines.")
print()
results = results.replace(np.nan, None)
results = tabulate.tabulate(results, headers="keys", tablefmt="github", floatfmt=".3f")
print(results)



from __future__ import annotations

import abc
from collections.abc import Iterable
from collections.abc import Sequence
import os

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as parquet

import audformat

from audb.core.api import dependencies
from audb.core.cache import database_cache_root
from audb.core.dependencies import error_message_missing_object
from audb.core.flavor import Flavor
from audb.core.load import _load_files
from audb.core.load import _misc_tables_used_in_table
from audb.core.load import _update_path
from audb.core.load import latest_version
from audb.core.load import load_header_to
from audb.core.load import load_media
from audb.core.lock import FolderLock


class DatabaseIterator(audformat.Database, metaclass=abc.ABCMeta):
    r"""Database iterator.

    This class cannot be created directly,
    but only by calling :func:`audb.stream`.

    Examples:
        Create :class:`audb.DatabaseIterator` object.

        >>> db = audb.stream(
        ...     "emodb",
        ...     "files",
        ...     version="1.4.1",
        ...     batch_size=4,
        ...     only_metadata=True,
        ...     full_path=False,
        ...     verbose=False,
        ... )

        The :class:`audb.DatabaseIterator` object
        is restricted to the requested table,
        and all related schemes
        and misc tables
        used as labels in a related scheme.

        >>> db
        name: emodb
        ...
        schemes:
          age: {description: Age of speaker, dtype: int, minimum: 0}
          duration: {dtype: time}
          gender:
            description: Gender of speaker
            dtype: str
            labels: [female, male]
          language: {description: Language of speaker, dtype: str}
          speaker: {description: The actors could produce each sentence as often as they liked
              and were asked to remember a real situation from their past when they had felt
              this emotion., dtype: int, labels: speaker}
          transcription:
            description: Sentence produced by actor.
            dtype: str
            labels: ...
        tables:
          files:
            type: filewise
            columns:
              duration: {scheme_id: duration}
              speaker: {scheme_id: speaker}
              transcription: {scheme_id: transcription}
        misc_tables:
          speaker:
            levels: {speaker: int}
            columns:
              age: {scheme_id: age}
              gender: {scheme_id: gender}
              language: {scheme_id: language}
        ...

        Request the first batch of data.

        >>> next(db)
                                         duration  speaker transcription
        file
        wav/03a01Fa.wav    0 days 00:00:01.898250        3           a01
        wav/03a01Nc.wav    0 days 00:00:01.611250        3           a01
        wav/03a01Wa.wav 0 days 00:00:01.877812500        3           a01
        wav/03a02Fc.wav    0 days 00:00:02.006250        3           a02

        During the iteration,
        the :class:`audb.DatabaseIterator` object
        provides access to the current batch of data.

        >>> db["files"].get(map={"speaker": "age"})
                                         duration transcription  age
        file
        wav/03a01Fa.wav    0 days 00:00:01.898250           a01   31
        wav/03a01Nc.wav    0 days 00:00:01.611250           a01   31
        wav/03a01Wa.wav 0 days 00:00:01.877812500           a01   31
        wav/03a02Fc.wav    0 days 00:00:02.006250           a02   31

    """  # noqa: E501

    def __init__(
        self,
        db: audformat.Database,
        table: str,
        *,
        version: str,
        map: dict[str, str | Sequence[str]],
        batch_size: int,
        shuffle: bool,
        buffer_size: int,
        only_metadata: bool,
        bit_depth: int,
        channels: int | Sequence[int],
        format: str,
        mixdown: bool,
        sampling_rate: int,
        full_path: bool,
        cache_root: str,
        num_workers: int | None,
        timeout: float,
        verbose: bool,
    ):
        self._cleanup_database(db, table)

        # Transfer attributes of database object
        for attr in db.__dict__.keys():
            setattr(self, attr, getattr(db, attr))

        self._table = table
        self._version = version
        self._map = map
        self._batch_size = batch_size
        self._shuffle = shuffle
        self._buffer_size = buffer_size
        self._only_metadata = only_metadata
        self._bit_depth = bit_depth
        self._channels = channels
        self._format = format
        self._mixdown = mixdown
        self._sampling_rate = sampling_rate
        self._full_path = full_path
        self._cache_root = cache_root
        self._num_workers = num_workers
        self._timeout = timeout
        self._verbose = verbose

        if shuffle:
            self._samples = buffer_size
        else:
            self._samples = batch_size
        self._buffer = pd.DataFrame()
        self._stream = self._initialize_stream()

    def __iter__(self) -> DatabaseIterator:
        r"""Iterator generator."""
        return self

    def __next__(self) -> pd.DataFrame:
        r"""Iterate database."""
        # Load part of table
        df = self._get_batch()
        self[self._table]._df = df

        # Load corresponding media files
        self._load_media(df)

        # Map column values
        if self._map is not None:
            df = self[self._table].get(map=self._map)

        # Adjust full paths and file extensions in table
        _update_path(
            self,
            self.root,
            self._full_path,
            self._format,
            self._num_workers,
            self._verbose,
        )

        return df

    @abc.abstractmethod
    def _initialize_stream(self) -> Iterable:
        r"""Create table iterator object.

        This method needs to be implemented
        for the table file types
        in the classes,
        that inherit from :class:`audb.DatabaseIterator`.

        Returns:
            table iterator

        """
        return  # pragma: nocover

    @staticmethod
    def _cleanup_database(db: audformat.Database, table: str):
        r"""Remove parts of database, not used by table.

        Args:
            db: database object
            table: table ID

        """
        tables = [table]
        misc_tables = _misc_tables_used_in_table(db[table])

        # Remove non-requested table
        db.drop_tables([table for table in list(db.tables) if table not in tables])

        # Remove unused schemes
        used_schemes = []
        for table in misc_tables + tables:
            for column_id, column in db[table].columns.items():
                if column.scheme_id is not None:
                    used_schemes.append(column.scheme_id)
        for scheme in list(db.schemes):
            if scheme not in used_schemes:
                del db.schemes[scheme]

        # Remove misc tables not required by the schemes of table
        db.drop_tables(
            [
                misc_table
                for misc_table in list(db.misc_tables)
                if misc_table not in misc_tables + tables
            ]
        )

        # Remove unused splits
        used_splits = [
            db[table].split_id for table in list(db) if db[table].split_id is not None
        ]
        for split in list(db.splits):
            if split not in used_splits:
                del db.splits[split]

    def _get_batch(self) -> pd.DataFrame:
        r"""Read table batch.

        Returns:
            dataframe

        """
        if self._shuffle:
            buffer_read_length = self._batch_size
            df1 = pd.DataFrame()
            if len(self._buffer) < self._batch_size:
                if len(self._buffer) > 0:
                    # Empty current buffer,
                    # before refilling
                    df1 = self._buffer
                    buffer_read_length = self._batch_size - len(self._buffer)
                self._buffer = self._read_dataframe()
                # Shuffle data
                self._buffer = self._buffer.sample(frac=1)

            df2 = self._buffer.iloc[:buffer_read_length, :]
            self._buffer.drop(index=df2.index, inplace=True)
            df = pd.concat([df1, df2])

        else:
            df = self._read_dataframe()

        if len(df) == 0:
            raise StopIteration

        return df

    def _load_media(self, df: pd.DataFrame):
        r"""Load media file for batch.

        Args:
            df: dataframe of batch

        """
        if audformat.is_segmented_index(df.index):
            media = list(df.index.get_level_values("file"))
        elif audformat.is_filewise_index(df.index):
            media = list(df.index)
        else:
            media = []
        if not self._only_metadata and len(media) > 0:
            load_media(
                self.name,
                media,
                version=self._version,
                bit_depth=self._bit_depth,
                channels=self._channels,
                format=self._format,
                mixdown=self._mixdown,
                sampling_rate=self._sampling_rate,
                cache_root=self._cache_root,
                num_workers=self._num_workers,
                timeout=self._timeout,
                verbose=self._verbose,
            )

    def _postprocess_batch(self, batch: object) -> pd.DataFrame:
        r"""Post-process batch data to desired dataframe.

        Args:
            batch: input data

        Returns:
            dataframe

        """
        return batch  # pragma: nocover

    def _postprocess_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        r"""Post-process dataframe to have correct index and data types.

        Args:
            df: dataframe

        Returns:
            dataframe

        """
        # Adjust dtypes and set index
        df = self[self._table]._pyarrow_convert_dtypes(df, convert_all=False)
        index_columns = list(self[self._table]._levels_and_dtypes.keys())
        df = self[self._table]._set_index(df, index_columns)
        return df

    def _read_dataframe(self) -> pd.DataFrame:
        r"""Read dataframe from table.

        Returns:
            dataframe

        """
        try:
            df = self._postprocess_dataframe(
                self._postprocess_batch(next(iter(self._stream)))
            )
        except StopIteration:
            # Ensure return an empty dataframe,
            # at the last iteration,
            # when no remaining data is left
            df = pd.DataFrame()
        return df


class DatabaseIteratorCsv(DatabaseIterator):
    def _initialize_stream(self):
        # Prepare settings for csv file reading

        # index
        columns_and_dtypes = self[self._table]._levels_and_dtypes.copy()
        # add columns
        for column_id, column in self[self._table].columns.items():
            if column.scheme_id is not None:
                columns_and_dtypes[column_id] = self.schemes[column.scheme_id].dtype
            else:
                columns_and_dtypes[column_id] = audformat.define.DataType.OBJECT

        # Replace data type with converter for dates or timestamps
        converters = {}
        dtypes_wo_converters = {}
        for column, dtype in columns_and_dtypes.items():
            if dtype == audformat.define.DataType.DATE:
                converters[column] = lambda x: pd.to_datetime(x)
            elif dtype == audformat.define.DataType.TIME:
                converters[column] = lambda x: pd.to_timedelta(x)
            else:
                dtypes_wo_converters[column] = audformat.core.common.to_pandas_dtype(
                    dtype
                )

        if self._samples == 0:
            # `pandas.read_csv()` does not support a `chunksize=0`
            return []
        else:
            file = os.path.join(self.root, f"db.{self._table}.csv")
            return pd.read_csv(
                file,
                chunksize=self._samples,
                usecols=list(columns_and_dtypes.keys()),
                dtype=dtypes_wo_converters,
                converters=converters,
                float_precision="round_trip",
            )


class DatabaseIteratorParquet(DatabaseIterator):
    def _initialize_stream(self) -> pa.RecordBatch:
        file = os.path.join(self.root, f"db.{self._table}.parquet")
        return parquet.ParquetFile(file).iter_batches(batch_size=self._samples)

    def _postprocess_batch(self, batch: pa.RecordBatch) -> pd.DataFrame:
        df = batch.to_pandas(
            deduplicate_objects=False,
            types_mapper={
                pa.string(): pd.StringDtype(),
            }.get,  # we have to provide a callable, not a dict
        )
        return df


def stream(
    name: str,
    table: str,
    *,
    version: str = None,
    map: dict[str, str | Sequence[str]] = None,
    batch_size: int = 16,
    shuffle: bool = False,
    buffer_size: int = 100_000,
    only_metadata: bool = False,
    bit_depth: int = None,
    channels: int | Sequence[int] = None,
    format: str = None,
    mixdown: bool = False,
    sampling_rate: int = None,
    full_path: bool = True,
    cache_root: str = None,
    num_workers: int | None = 1,
    timeout: float = -1,
    verbose: bool = True,
) -> DatabaseIterator:
    r"""Stream table and media files of a database.

    Loads only the first ``batch_size`` rows
    of a table into memory,
    and downloads only the related media files,
    if any media files are requested.

    By setting
    ``bit_depth``,
    ``channels``,
    ``format``,
    ``mixdown``,
    and ``sampling_rate``
    we can request a specific flavor of the database.
    In that case media files are automatically converted to the desired
    properties (see also :class:`audb.Flavor`).

    Args:
        name: name of database
        table: name of table
        version: version string, latest if ``None``
        map: map scheme or scheme fields to column values.
            For example if your table holds a column ``speaker`` with
            speaker IDs, which is assigned to a scheme that contains a
            dict mapping speaker IDs to age and gender entries,
            ``map={'speaker': ['age', 'gender']}``
            will replace the column with two new columns that map ID
            values to age and gender, respectively.
            To also keep the original column with speaker IDS, you can do
            ``map={'speaker': ['speaker', 'age', 'gender']}``
        batch_size: number of table rows
            to return in one iteration
        shuffle: if ``True``,
            it first reads ``buffer_size`` rows from the table
            and selects ``batch_size`` randomly from them
        buffer_size: number of table rows
            to be loaded
            when ``shuffle`` is ``True``
        only_metadata: load only header and tables of database
        bit_depth: bit depth, one of ``16``, ``24``, ``32``
        channels: channel selection, see :func:`audresample.remix`.
            Note that media files with too few channels
            will be first upsampled by repeating the existing channels.
            E.g. ``channels=[0, 1]`` upsamples all mono files to stereo,
            and ``channels=[1]`` returns the second channel
            of all multi-channel files
            and all mono files
        format: file format, one of ``'flac'``, ``'wav'``
        mixdown: apply mono mix-down
        sampling_rate: sampling rate in Hz, one of
            ``8000``, ``16000``, ``22050``, ``24000``, ``44100``, ``48000``
        full_path: replace relative with absolute file paths
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used
        num_workers: number of parallel jobs or 1 for sequential
            processing. If ``None`` will be set to the number of
            processors on the machine multiplied by 5
        timeout: maximum wait time if another thread or process is already
            accessing the database. If timeout is reached, ``None`` is
            returned. If timeout < 0 the method will block until the
            database can be accessed
        verbose: show debug messages

    Returns:
        database object

    Raises:
        ValueError: if table is requested
            that is not part of the database
        ValueError: if a non-supported ``bit_depth``,
            ``format``,
            or ``sampling_rate``
            is requested
        RuntimeError: if a flavor is requested,
            but the database contains media files,
            that don't contain audio,
            e.g. text files

    Examples:
        >>> import numpy as np
        >>> np.random.seed(1)
        >>> db = audb.stream(
        ...     "emodb",
        ...     "files",
        ...     version="1.4.1",
        ...     batch_size=4,
        ...     shuffle=True,
        ...     only_metadata=True,
        ...     full_path=False,
        ...     verbose=False,
        ... )
        >>> next(db)
                                         duration  speaker transcription
        file
        wav/14a05Fb.wav 0 days 00:00:03.128687500       14           a05
        wav/15a05Eb.wav 0 days 00:00:03.993562500       15           a05
        wav/12a05Nd.wav    0 days 00:00:03.185875       12           a05
        wav/13a07Na.wav 0 days 00:00:01.911687500       13           a07

    """
    if version is None:
        version = latest_version(name)

    # Extract kwargs
    # to pass on to the DatabaseIterator constructor
    kwargs = {k: v for (k, v) in locals().items() if k not in ["name", "table"]}

    flavor = Flavor(
        bit_depth=bit_depth,
        channels=channels,
        format=format,
        mixdown=mixdown,
        sampling_rate=sampling_rate,
    )
    db_root = database_cache_root(name, version, cache_root, flavor)

    deps = dependencies(
        name,
        version=version,
        cache_root=cache_root,
        verbose=verbose,
    )

    if table not in deps.table_ids:
        msg = error_message_missing_object("table", [table], name, version)
        raise ValueError(msg)

    with FolderLock(db_root):
        # Start with database header without tables
        db, backend_interface = load_header_to(
            db_root,
            name,
            version,
            flavor=flavor,
            add_audb_meta=True,
        )

        # Misc tables required by schemes of requested table
        misc_tables = _misc_tables_used_in_table(db[table])

        # Load table files
        _load_files(
            misc_tables + [table],
            "table",
            backend_interface,
            db_root,
            db,
            version,
            None,
            deps,
            Flavor(),
            cache_root,
            False,  # pickle_tables
            num_workers,
            verbose,
        )

        # Load misc tables completely
        for misc_table in misc_tables:
            table_file = os.path.join(db_root, f"db.{misc_table}")
            db[misc_table].load(table_file)

    if os.path.exists(os.path.join(db_root, f"db.{table}.parquet")):
        return DatabaseIteratorParquet(db, table, **kwargs)
    else:
        return DatabaseIteratorCsv(db, table, **kwargs)



from datetime import date

import toml

import audeer

import audb


config = toml.load(audeer.path("..", "pyproject.toml"))


# Project -----------------------------------------------------------------
project = config["project"]["name"]
copyright = f"2020-{date.today().year} audEERING GmbH"
author = ", ".join(author["name"] for author in config["project"]["authors"])
version = audeer.git_repo_version()
title = "Documentation"


# General -----------------------------------------------------------------
master_doc = "index"
source_suffix = ".rst"
templates_path = ["_templates"]
exclude_patterns = [
    "api-src",
    "build",
    "tests",
    "Thumbs.db",
    ".DS_Store",
]
pygments_style = None
extensions = [
    "sphinx.ext.graphviz",
    "sphinx.ext.napoleon",  # support for Google-style docstrings
    "sphinx_autodoc_typehints",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx_copybutton",
    "sphinx_apipages",
]

napoleon_use_ivar = True  # List of class attributes
# autodoc_inherit_docstrings = False  # disable docstring inheritance
intersphinx_mapping = {
    "audbackend": ("https://audeering.github.io/audbackend/", None),
    "audeer": ("https://audeering.github.io/audeer/", None),
    "audformat": ("http://audeering.github.io/audformat/", None),
    "audiofile": ("https://audeering.github.io/audiofile/", None),
    "audobject": ("https://audeering.github.io/audobject/", None),
    "audresample": ("https://audeering.github.io/audresample/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "python": ("https://docs.python.org/3/", None),
}
linkcheck_ignore = [
    "https://gitlab.audeering.com",
    (  # This is a permalink, which provides a false-positive result
        "https://github.com/audeering/audbackend/blob/"
        "edd23462799ae9052a43cdd045698f78e19dbcaf"
    ),
    "http://emodb.bilderbar.info/start.html",
    # This permalink fails from time to time and should never be broken
    "https://superuser.com/a/264406",
]
# Ignore package dependencies during building the docs
# This fixes URL link issues with pandas and sphinx_autodoc_typehints
autodoc_mock_imports = [
    "pandas",
]
graphviz_output_format = "svg"

apipages_hidden_methods = [
    "__call__",
    "__contains__",
    "__eq__",
    "__getitem__",
    "__iter__",
    "__len__",
    "__next__",
]

# HTML --------------------------------------------------------------------
html_theme = "sphinx_audeering_theme"
html_theme_options = {
    "display_version": True,
    "logo_only": False,
    "wide_pages": ["data-example"],
    "footer_links": False,
}
html_context = {
    "display_github": True,
}
html_title = title


# Cache databases to avoid progress bar in code examples ------------------
audb.config.REPOSITORIES = [
    audb.Repository(
        name="data-public",
        host="https://audeering.jfrog.io/artifactory",
        backend="artifactory",
    )
]
database_name = "emodb"
database_version = "1.4.1"
if not audb.exists(database_name, version=database_version):
    print(f"Pre-caching {database_name} v{database_version}")
    audb.load(
        database_name,
        version=database_version,
        num_workers=5,
        only_metadata=True,
        verbose=False,
    )
if not audb.exists(
    database_name,
    version=database_version,
    format="flac",
    sampling_rate=44100,
):
    print(f"Pre-caching {database_name} v{database_version} {{flac, 44100Hz}}")
    audb.load(
        database_name,
        version=database_version,
        format="flac",
        sampling_rate=44100,
        num_workers=5,
        only_metadata=True,
        verbose=False,
    )



from doctest import ELLIPSIS
from doctest import NORMALIZE_WHITESPACE
import os

import pytest
import sybil
from sybil.parsers.rest import DocTestParser
from sybil.parsers.rest import PythonCodeBlockParser
from sybil.parsers.rest import SkipParser

import audb
from audb.core.conftest import cache  # noqa: F401
from audb.core.conftest import imports
from audb.core.conftest import public_repository  # noqa: F401


class DefaultConfiguration:
    """Context manager to provide default configuration values."""

    def __init__(self):
        self.config = audb.core.config.load_configuration_file(
            audb.core.config.global_config_file
        )
        self._saved_state = {}

    def __enter__(self):
        """Save current state."""
        self._saved_state = {
            "cache_root": audb.config.CACHE_ROOT,
            "shared_cache_root": audb.config.SHARED_CACHE_ROOT,
            "repositories": audb.config.REPOSITORIES,
        }
        # Set default values
        audb.config.CACHE_ROOT = self.config["cache_root"]
        audb.config.SHARED_CACHE_ROOT = self.config["shared_cache_root"]
        audb.config.REPOSITORIES = [
            audb.Repository(repo["name"], repo["host"], repo["backend"])
            for repo in self.config["repositories"]
        ]

    def __exit__(self, *args):
        """Restore previous state."""
        for key, value in self._saved_state.items():
            setattr(audb.config, key, value)


@pytest.fixture(scope="module")
def default_configuration():
    """Set config values to default values."""
    with DefaultConfiguration():
        yield


@pytest.fixture(scope="module")
def run_in_tmpdir(tmpdir_factory):
    """Move to a persistent tmpdir for execution of a whole file."""
    tmpdir = tmpdir_factory.mktemp("tmp")
    current_dir = os.getcwd()
    os.chdir(tmpdir)

    yield

    os.chdir(current_dir)


# Collect doctests
#
# We use several `sybil.Sybil` instances
# to pass different fixtures for different files
#
parsers = [
    DocTestParser(optionflags=ELLIPSIS + NORMALIZE_WHITESPACE),
    PythonCodeBlockParser(),
    SkipParser(),
]
pytest_collect_file = sybil.sybil.SybilCollection(
    (
        sybil.Sybil(
            parsers=parsers,
            filenames=[
                "authentication.rst",
                "overview.rst",
                "quickstart.rst",
                "dependencies.rst",
                "load.rst",
                "audb.info.rst",
            ],
            fixtures=[
                "cache",
                "run_in_tmpdir",
                "public_repository",
            ],
            setup=imports,
        ),
        sybil.Sybil(
            parsers=parsers,
            filenames=["publish.rst"],
            fixtures=["cache", "run_in_tmpdir"],
            setup=imports,
        ),
        sybil.Sybil(
            parsers=parsers,
            filenames=["configuration.rst", "caching.rst"],
            fixtures=["default_configuration"],
            setup=imports,
        ),
    )
).pytest()



import sys

import audbackend


class Repository:
    r"""Repository object.

    It stores all information
    needed to address a repository:
    the repository name,
    host,
    and the backend name.
    With :meth:`Repository.create_backend_interface`
    it also provides a method
    to create a backend interface
    to access the repository.

    Args:
        name: repository name
        host: repository host
        backend: repository backend

    Examples:
        >>> Repository("data-local", "/data", "file-system")
        Repository('data-local', '/data', 'file-system')

    """

    _backends = {
        "file-system": audbackend.backend.FileSystem,
        "minio": audbackend.backend.Minio,
        "s3": audbackend.backend.Minio,
    }

    if hasattr(audbackend.backend, "Artifactory"):
        _backends["artifactory"] = audbackend.backend.Artifactory  # pragma: no cover

    backend_registry = _backends
    r"""Backend registry.

    Holds mapping between registered backend names,
    and their corresponding backend classes.
    The ``"artifactory"`` backend is currently not available
    under Python >=3.12.

    """

    def __init__(
        self,
        name: str,
        host: str,
        backend: str,
    ):
        self.name = name
        r"""Repository name."""
        self.host = host
        r"""Repository host."""
        self.backend = backend
        r"""Repository backend name."""

    def __eq__(self, other) -> bool:
        """Compare two repository instances.

        Args:
            other: repository instance

        Returns:
            ``True`` if the string representation of the repositories matches

        """
        return str(self) == str(other)

    def __repr__(self):  # noqa: D105
        return (
            f"Repository("
            f"'{self.name}', "
            f"'{self.host}', "
            f"'{self.backend}'"
            f")"
        )

    def create_backend_interface(self) -> type[audbackend.interface.Base]:
        r"""Create backend interface to access repository.

        It wraps an :class:`audbackend.interface.Versioned` interface
        around it.
        The files will then be stored
        with the following structure on the backend
        (shown by the example of version 1.0.0 of the emodb dataset)::

            emodb/1.0.0/db.yaml            <-- header
            emodb/1.0.0/db.zip             <-- dependency table
            emodb/attachment/1.0.0/...     <-- attachments
            emodb/media/1.0.0/...          <-- media files
            emodb/meta/1.0.0/...           <-- tables

        When :attr:`Repository.backend` equals ``artifactory``,
        it wraps an :class:`audbackend.interface.Maven` interface
        around it.
        The files will then be stored
        with the following structure on the Artifactory backend
        (shown by the example of version 1.0.0 of the emodb dataset)::

            emodb/db/1.0.0/db-1.0.0.yaml   <-- header
            emodb/db/1.0.0/db-1.0.0.zip    <-- dependency table
            emodb/attachment/.../1.0.0/... <-- attachments
            emodb/media/.../1.0.0/...      <-- media files
            emodb/meta/.../1.0.0/...       <-- tables

        The returned backend instance
        has not yet established a connection to the backend.
        To establish a connection,
        use the backend with a ``with`` statement,
        or use the ``open()`` and ``close()`` methods of the backend class.
        The backend is stored as the inside the ``backend`` attribute
        of the returned backend interface.

        Returns:
            interface to repository

        Raises:
            ValueError: if an artifactory backend is requested in Python>=3.12
            ValueError: if a non-supported backend is requested

        """
        if sys.version_info >= (3, 12) and self.backend == "artifactory":
            raise ValueError(  # pragma: no cover
                "The 'artifactory' backend is not supported in Python>=3.12"
            )
        if self.backend not in self.backend_registry:
            raise ValueError(f"'{self.backend}' is not a registered backend")
        backend_class = self.backend_registry[self.backend]
        backend = backend_class(self.host, self.name)
        if self.backend == "artifactory":
            interface = audbackend.interface.Maven(backend)  # pragma: no cover
        else:
            interface = audbackend.interface.Versioned(backend)
        return interface

    @classmethod
    def register(
        cls,
        backend_name: str,
        backend_class: type[audbackend.backend.Base],
    ):
        r"""Register backend class.

        Adds an entry to the dictionary
        stored in the class variable :data:`Repository.backend_registry`,
        mapping a backend name
        to an actual backend class.

        Args:
            backend_name: name of the backend,
                e.g. ``"file-system"``
            backend_class: class of the backend,
                that should be associated with ``backend_name``,
                e.g. ``"audbackend.backend.Filesystem"``

        Examples:
            >>> import audbackend
            >>> audb.Repository.register("file-system", audbackend.backend.FileSystem)

        """
        cls.backend_registry[backend_name] = backend_class



from __future__ import annotations

from collections.abc import Sequence
import os
import tempfile

import pandas as pd

import audbackend
import audeer
import audformat

from audb.core import define
from audb.core import utils
from audb.core.cache import database_cache_root
from audb.core.cache import default_cache_root
from audb.core.config import config
from audb.core.dependencies import Dependencies
from audb.core.dependencies import download_dependencies
from audb.core.dependencies import upload_dependencies
from audb.core.flavor import Flavor
from audb.core.lock import FolderLock
from audb.core.repository import Repository


def available(
    *,
    only_latest: bool = False,
) -> pd.DataFrame:
    r"""List all databases that are available to the user.

    Args:
        only_latest: include only latest version of database

    Returns:
        table with database name as index,
        and backend, host, repository, version as columns

    Examples:
        >>> df = audb.available(only_latest=True)
        >>> df.loc[["air", "emodb"]]
              backend                                   host   repository version
        name
        air        s3  s3.dualstack.eu-north-1.amazonaws.com  audb-public   1.4.2
        emodb      s3  s3.dualstack.eu-north-1.amazonaws.com  audb-public   1.4.1

    """  # noqa: E501
    databases = []

    def add_database(name: str, version: str, repository: Repository):
        """Add database to list of databases."""
        databases.append(
            [
                name,
                repository.backend,
                repository.host,
                repository.name,
                version,
            ]
        )

    for repository in config.REPOSITORIES:
        try:
            backend_interface = repository.create_backend_interface()
            with backend_interface.backend as backend:
                if repository.backend == "artifactory":  # pragma: nocover
                    # avoid backend_interface.ls('/')
                    # which is very slow on Artifactory
                    # see https://github.com/audeering/audbackend/issues/132
                    for p in backend.path("/"):
                        name = p.name
                        try:
                            for version in [str(x).split("/")[-1] for x in p / "db"]:
                                add_database(name, version, repository)
                        except FileNotFoundError:
                            # If the `db` folder does not exist,
                            # we do not include the dataset
                            pass

                elif repository.backend in ["minio", "s3"]:
                    # Avoid `ls(recursive=True)` for S3 and MinIO
                    # as this is slow for large databases
                    for obj in backend._client.list_objects(repository.name):
                        name = obj.object_name[:-1]  # remove "/" at end
                        header_file = f"/{name}/{define.HEADER_FILE}"
                        for _obj in backend._client.list_objects(
                            repository.name, f"{name}/"
                        ):
                            version = _obj.object_name.split("/")[1]
                            header_file = f"/{name}/{version}/{define.HEADER_FILE}"
                            if version not in [
                                "attachment",
                                "media",
                                "meta",
                            ] and backend.exists(header_file):
                                add_database(name, version, repository)

                else:
                    for path, version in backend_interface.ls("/"):
                        if path.endswith(define.HEADER_FILE):
                            name = path.split("/")[1]
                            add_database(name, version, repository)

        except (audbackend.BackendError, ValueError):
            continue

    df = pd.DataFrame.from_records(
        databases,
        columns=["name", "backend", "host", "repository", "version"],
    )
    if only_latest:
        # Pick latest version for every database, see
        # https://stackoverflow.com/a/53842408
        df = df[
            df["version"]
            == df.groupby("name")["version"].transform(
                lambda x: audeer.sort_versions(x)[-1]
            )
        ]
    else:
        # Sort by version
        df = df.sort_values(by=["version"], key=audeer.sort_versions)
    df = df.sort_values(by=["name"])
    return df.set_index("name")


def cached(
    cache_root: str = None,
    *,
    name: str = None,
    shared: bool = False,
) -> pd.DataFrame:
    r"""List available databases and flavors in the cache.

    Args:
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used
        name: name of database.
            If provided,
            it will show only cached versions of that database
        shared: include databases from shared cache

    Returns:
        cached databases
        with cache path as index,
        and name,
        flavor_id,
        version,
        complete,
        bit_depth,
        channels,
        format,
        mixdown,
        sampling_rate
        as columns

    Examples:
        >>> db = audb.load(
        ...     "emodb",
        ...     version="1.4.1",
        ...     only_metadata=True,
        ...     full_path=False,
        ...     verbose=False,
        ... )
        >>> df = audb.cached()
        >>> print(df.iloc[0].to_string())
        name                emodb
        flavor_id        d3b62a9b
        version             1.4.1
        complete            False
        bit_depth            None
        channels             None
        format               None
        mixdown             False
        sampling_rate        None

    """  # noqa: E501
    cache_root = audeer.path(cache_root or default_cache_root(shared=shared))

    columns = [
        "name",
        "flavor_id",
        "version",
        "complete",
        "bit_depth",
        "channels",
        "format",
        "mixdown",
        "sampling_rate",
    ]
    df = pd.DataFrame([], columns=columns)

    if not os.path.exists(cache_root):
        return df

    database_paths = audeer.list_dir_names(cache_root)
    for database_path in database_paths:
        database = os.path.basename(database_path)

        # Limit to databases of given name
        if name is not None and database != name:
            continue

        version_paths = audeer.list_dir_names(database_path)
        for version_path in version_paths:
            version = os.path.basename(version_path)

            # Skip tmp folder (e.g. 1.0.0~)
            if version.endswith("~"):  # pragma: no cover
                continue

            flavor_id_paths = audeer.list_dir_names(version_path)

            # Skip old audb cache (e.g. 1 as flavor)
            files = audeer.list_file_names(version_path, basenames=True)
            if (
                define.DEPENDENCY_FILE not in files
                and define.LEGACY_DEPENDENCY_FILE not in files
                and define.CACHED_DEPENDENCY_FILE not in files
            ):
                # Skip all cache entries
                # that don't contain a dependency file
                # as those stem from audb<1.0.0.
                continue  # pragma: no cover

            for flavor_id_path in flavor_id_paths:
                flavor_id = os.path.basename(flavor_id_path)
                files = audeer.list_file_names(flavor_id_path)
                files = [os.path.basename(f) for f in files]

                if define.HEADER_FILE in files:
                    db = audformat.Database.load(
                        flavor_id_path,
                        load_data=False,
                    )
                    flavor = db.meta["audb"]["flavor"]
                    complete = db.meta["audb"]["complete"]
                    df.loc[flavor_id_path] = [
                        database,
                        flavor_id,
                        version,
                        complete,
                        flavor["bit_depth"],
                        flavor["channels"],
                        flavor["format"],
                        flavor["mixdown"],
                        flavor["sampling_rate"],
                    ]

    # Replace NaN with None
    return df.where(pd.notnull(df), None)


def dependencies(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
    verbose: bool = False,
) -> Dependencies:
    r"""Database dependencies.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used
        verbose: show debug messages

    Returns:
        dependency object

    Examples:
        >>> deps = audb.dependencies("emodb", version="1.4.1")
        >>> deps.version("db.emotion.csv")
        '1.1.0'

    """
    if version is None:
        version = latest_version(name)

    db_root = database_cache_root(
        name,
        version,
        cache_root=cache_root,
    )
    cached_deps_file = os.path.join(db_root, define.CACHED_DEPENDENCY_FILE)

    with FolderLock(db_root):
        try:
            deps = Dependencies()
            deps.load(cached_deps_file)
        except Exception:  # does not catch KeyboardInterupt
            # If loading cached file fails, load again from backend
            backend_interface = utils.lookup_backend(name, version)
            deps = download_dependencies(backend_interface, name, version, verbose)
            # Store as pickle in cache
            deps.save(cached_deps_file)

    return deps


def exists(
    name: str,
    *,
    version: str = None,
    bit_depth: int = None,
    channels: int | Sequence[int] = None,
    format: str = None,
    mixdown: bool = False,
    sampling_rate: int = None,
    cache_root: str = None,
) -> bool:
    r"""Check if specified database flavor exists in local cache folder.

    Does not check for any flavor of the requested database in the cache,
    but only for a particular flavor.
    Note, that using only the name, e.g. ``audb.exists('emodb')``
    is also a single flavor.

    To list all available flavors of a particular database, use:

    .. code-block::

        audb.cached(name='emodb')

    Args:
        name: name of database
        version: version string, latest if ``None``
        bit_depth: bit depth, one of ``16``, ``24``, ``32``
        channels: channel selection, see :func:`audresample.remix`
        format: file format, one of ``'flac'``, ``'wav'``
        mixdown: apply mono mix-down
        sampling_rate: sampling rate in Hz, one of
            ``8000``, ``16000``, ``22050``, ``24000``, ``44100``, ``48000``
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        ``True`` if database flavor exists

    Examples:
        >>> db = audb.load(
        ...     "emodb",
        ...     version="1.4.1",
        ...     only_metadata=True,
        ...     verbose=False,
        ... )
        >>> audb.exists("emodb", version="1.4.1")
        True
        >>> audb.exists("emodb", version="1.4.1", format="wav")
        False

    """
    if version is None:
        version = latest_version(name)

    relative_flavor_path = flavor_path(
        name,
        version,
        channels=channels,
        format=format,
        mixdown=mixdown,
        bit_depth=bit_depth,
        sampling_rate=sampling_rate,
    )

    cache_roots = (
        [
            default_cache_root(True),  # check shared cache first
            default_cache_root(False),
        ]
        if cache_root is None
        else [audeer.path(cache_root, follow_symlink=True)]
    )
    for cache_root in cache_roots:
        db_root = os.path.join(cache_root, relative_flavor_path)
        if os.path.exists(db_root):
            return True

    return False


def flavor_path(
    name: str,
    version: str,
    *,
    bit_depth: int = None,
    channels: int | Sequence[int] = None,
    format: str = None,
    mixdown: bool = False,
    sampling_rate: int = None,
) -> str:
    r"""Flavor cache path.

    Returns the path under which :func:`audb.load` stores a specific
    flavor of a database in the cache folder, that is:

     ``<name>/<version>/<flavor-id>/``

    Note that the returned path is relative.
    To get the absolute path, do:

    .. code-block::

        os.path.join(
            audb.default_cache_root(...),
            audb.flavor_path(...),
        )

    Args:
        name: name of database
        version: version string
        bit_depth: bit depth, one of ``16``, ``24``, ``32``
        channels: channel selection, see :func:`audresample.remix`
        format: file format, one of ``'flac'``, ``'wav'``
        mixdown: apply mono mix-down
        sampling_rate: sampling rate in Hz, one of
            ``8000``, ``16000``, ``22050``, ``24000``, ``44100``, ``48000``

    Returns:
        flavor path relative to cache folder

    Raises:
        ValueError: if a non-supported ``bit_depth``,
            ``format``,
            or ``sampling_rate``
            is requested

    Examples:
        >>> audb.flavor_path("emodb", version="1.4.1").split(os.path.sep)
        ['emodb', '1.4.1', 'd3b62a9b']

    """
    flavor = Flavor(
        channels=channels,
        format=format,
        mixdown=mixdown,
        bit_depth=bit_depth,
        sampling_rate=sampling_rate,
    )

    return flavor.path(name, version)


def latest_version(
    name,
) -> str:
    r"""Latest version of database.

    Args:
        name: name of database

    Returns:
        version string

    Raises:
        RuntimeError: if no version exists for the requested database

    Examples:
        >>> audb.latest_version("emodb")
        '1.4.1'

    """
    vs = versions(name)
    if not vs:
        raise RuntimeError(
            f"Cannot find a version for database '{name}'.",
        )
    return vs[-1]


def remove_media(
    name: str,
    files: str | Sequence[str],
    *,
    verbose: bool = False,
):
    r"""Remove media from all versions.

    Be careful,
    this removes files from all published versions
    on all backends.
    Those files cannot be restored afterwards.

    Args:
        name: name of database
        files: list of files that should be removed
        verbose: show debug messages

    """
    if isinstance(files, str):
        files = [files]

    for version in versions(name):
        backend_interface = utils.lookup_backend(name, version)
        deps = download_dependencies(backend_interface, name, version, verbose)

        with tempfile.TemporaryDirectory() as db_root:
            # Track if we need to upload the dependency table again
            upload = False

            for file in audeer.progress_bar(
                files,
                disable=not verbose,
                desc=f"Remove media from v{version}",
            ):
                if file in deps.media:
                    archive = deps.archive(file)

                    # if archive exists in this version,
                    # remove file from it and re-publish
                    remote_archive = backend_interface.join(
                        "/", name, "media", archive + ".zip"
                    )
                    if backend_interface.exists(remote_archive, version):
                        files_in_archive = backend_interface.get_archive(
                            remote_archive,
                            db_root,
                            version,
                        )

                        if os.name == "nt":  # pragma: no cover
                            files_in_archive = [
                                file.replace(os.path.sep, "/")
                                for file in files_in_archive
                            ]

                        # skip if file was already deleted
                        if file in files_in_archive:
                            os.remove(os.path.join(db_root, file))
                            files_in_archive.remove(file)
                            backend_interface.put_archive(
                                db_root,
                                remote_archive,
                                version,
                                files=files_in_archive,
                            )

                    # mark file as removed
                    deps._remove(file)
                    upload = True

            # upload dependencies
            if upload:
                upload_dependencies(backend_interface, deps, db_root, name, version)


def repository(
    name: str,
    version: str,
) -> Repository:
    r"""Return repository that stores the requested database.

    If the database is stored in several repositories,
    only the first one is returned.
    The order of the repositories to look for the database
    is given by :attr:`config.REPOSITORIES`.

    Args:
        name: database name
        version: version string

    Returns:
        repository that contains the database

    Raises:
        RuntimeError: if database or version is not found

    Examples:
        >>> audb.repository("emodb", "1.4.1")
        Repository('audb-public', 's3.dualstack.eu-north-1.amazonaws.com', 's3')

    """  # noqa: E501
    if not versions(name):
        raise RuntimeError(f"Cannot find database " f"'{name}'.")
    return utils._lookup(name, version)[0]


def versions(
    name: str,
) -> list[str]:
    r"""Available versions of database.

    Args:
        name: name of database

    Returns:
        list of versions

    Examples:
        >>> audb.versions("emodb")
        ['1.1.0', '1.1.1', '1.2.0', '1.3.0', '1.4.0', '1.4.1']

    """
    vs = []
    for repository in config.REPOSITORIES:
        try:
            backend_interface = repository.create_backend_interface()
            with backend_interface.backend as backend:
                if repository.backend == "artifactory":  # pragma: nocover
                    # Do not use `backend_interface.versions()` on Artifactory,
                    # as calling `backend_interface.ls()` is slow on Artifactory,
                    # see https://github.com/devopshq/artifactory/issues/423.
                    # Instead, use `backend.path()`
                    # from the low level backend object
                    # to return an ArtifactoryPath object,
                    # which allows to walk through the dataset directory
                    # and to collect all existing versions.
                    folder = backend.join("/", name, "db")
                    path = backend.path(folder)
                    try:
                        if path.exists():
                            for p in path:
                                version = p.parts[-1]
                                header = p.joinpath(f"db-{version}.yaml")
                                if header.exists():
                                    vs.extend([version])
                    except Exception:
                        # Might happen,
                        # if a database is not available on the backend,
                        # or we don't have read permissions.
                        #
                        # We cannot test this at the moment
                        # on the public Artifactory server.
                        # Because after trying
                        # to connect to a path without read access
                        # the connection is also blocked for valid paths.
                        pass
                else:
                    header = backend_interface.join("/", name, "db.yaml")
                    vs.extend(
                        backend_interface.versions(
                            header,
                            suppress_backend_errors=True,
                        )
                    )
        except (audbackend.BackendError, ValueError):
            # If the backend cannot be accessed,
            # e.g. host or repository do not exist,
            # we skip it
            # and continue with the next repository
            continue
    return audeer.sort_versions(vs)



from __future__ import annotations

from collections.abc import Callable
from collections.abc import Sequence
import errno
import os
import re
import tempfile

import pandas as pd
import polars as pl
import pyarrow as pa
import pyarrow.csv as csv
import pyarrow.parquet as parquet

import audbackend
import audeer

from audb.core import define


pl.Config.set_tbl_hide_dataframe_shape(True)
pl.Config.set_tbl_formatting("NOTHING")
pl.Config.set_tbl_hide_column_data_types(True)


class Dependencies:
    r"""Dependencies of a database.

    :class:`audb.Dependencies` gathers
    all database media, table, and attachment files
    and metadata about them
    in a single object.
    The metadata contains information
    about the single files
    like duration,
    but also what version of the file is required.

    The dependencies of a database can be requested with
    :func:`audb.dependencies`.

    Examples:
        >>> deps = Dependencies()
        >>> deps()
        Empty DataFrame
        Columns: [archive, bit_depth, channels, checksum, duration, format, removed, sampling_rate, type, version]
        Index: []
        >>> # Request dependencies for emodb 1.4.1
        >>> deps = audb.dependencies("emodb", version="1.4.1")
        >>> # List all files or archives
        >>> deps.files[:3]
        ['db.emotion.csv', 'db.files.csv', 'wav/03a01Fa.wav']
        >>> deps.archives[:2]
        ['005d2b91-5317-0c80-d602-6d55f0323f8c', '014f82d8-3491-fd00-7397-c3b2ac3b2875']
        >>> # Access properties for a given file
        >>> deps.archive("wav/03a01Fa.wav")
        'c1f5cc6f-6d00-348a-ba3b-4adaa2436aad'
        >>> deps.duration("wav/03a01Fa.wav")
        1.89825
        >>> deps.removed("wav/03a01Fa.wav")
        False
        >>> # Check if a file is part of the dependencies
        >>> "wav/03a01Fa.wav" in deps
        True

    """  # noqa: E501

    def __init__(self):
        data = {}
        for name, dtype in zip(
            define.DEPEND_FIELD_NAMES.values(),
            define.DEPEND_FIELD_DTYPES.values(),
        ):
            # data[name] = pd.Series(dtype=dtype)
            data[name] = pl.Series(dtype=dtype)

        # self._df = pd.DataFrame(data)

        # always an empty frame.
        # deps can only be loaded or self._df set
        self._df = pl.DataFrame(data)
        self.index_col = define.DEPEND_INDEX_COLNAME
        # self._df.index = self._df.index.astype(define.DEPEND_INDEX_DTYPE)

        # pyarrow schema
        # used for reading and writing files
        # self._schema = pa.schema(
        #     [
        #         ("file", pa.string()),
        #         ("archive", pa.string()),
        #         ("bit_depth", pa.int32()),
        #         ("channels", pa.int32()),
        #         ("checksum", pa.string()),
        #         ("duration", pa.float64()),
        #         ("format", pa.string()),
        #         ("removed", pa.int32()),
        #         ("sampling_rate", pa.int32()),
        #         ("type", pa.int32()),
        #         ("version", pa.string()),
        #     ]
        # )

        # self._schema = utils.pascheme_to_plscheme(self._schema)
        # polars the df has a schema, it can be converted to dict
        # using dict(self._df.schema)
        self._schema = self._df.schema

    def __call__(self) -> pd.DataFrame:
        r"""Return dependencies as a table.

        Returns:
            table with dependencies

        """
        return self._df

    def __contains__(self, file: str) -> bool:
        r"""Check if file is part of dependencies.

        Args:
            file: relative file path

        Returns:
            ``True`` if a dependency to the file exists

        """
        return file in self._idx

    def __eq__(self, other: "Dependencies") -> bool:
        r"""Check if two dependency tables are equal.

        Args:
            other: dependency table to compare against

        Returns:
            ``True`` if both dependency tables have the same entries

        """
        return self._df.equals(other._df)

    def __getitem__(self, file: str) -> list:
        r"""File information.

        Args:
            file: relative file path

        Returns:
            list with meta information


        """
        # (self._df[self.index_col] == file).to_list()
        # return .loc[file].tolist()
        # item = pl.Series(self._df.filter(pl.col(self.index_col) == file)).to_list()
        # even slower:
        # item = np.array(self._df.filter(pl.col(self.index_col) == file)).tolist()[0]
        # pandas:
        # self._df.loc[file].tolist()
        item = list(self._df.row(self._idx[file]))[1:]
        return item

    def __len__(self) -> int:
        r"""Number of all media, table, attachment files."""
        return len(self._idx)

    def __str__(self) -> str:  # noqa: D105
        n = 15  # number of lines identical to pandas
        str_repr = str(self._df.head(n))
        return str_repr

    @property
    def archives(self) -> list[str]:
        r"""All media, table, attachment archives.

        Return:
            list of archives


        This is not benchmarked at the moment
        Current implementation is the fastest of

        self._df["archive"].sort(descending=False).unique(maintain_order=True).to_list()
        self._df["archive"].unique(maintain_order=False).sort(descending=False).to_list()
        0.144546
        sorted(self._df["archive"].unique().to_list())

        Sorting?

        Result:
        ['archive-0', 'archive-1', 'archive-2', 'archive-3', ...]
        """
        # 0.014
        # return (
        #     self._df["archive"]
        #     .unique(maintain_order=False)
        #     .sort(descending=False)
        #     .to_list()
        # )

        # 0.014
        # return (
        #     self._df.lazy()
        #     .select(pl.col("archive"))
        #     .unique()
        #     .sort(pl.col("archive"))
        #     .collect()["archive"]
        #     .to_list()
        # )

        # 0.014
        return pl.Series(self._df.select(pl.col("archive"))).unique().sort().to_list()

        # 0.413 - very slow to use list sorting
        # return sorted(
        #     pl.Series(self._df.select(pl.col("archive"))).unique().to_list()
        # )

    @property
    def attachments(self) -> list[str]:
        r"""Attachment paths (can be a file or a folder).

        Returns:
            list of attachments

        """
        deptype = define.DependType.ATTACHMENT
        return self._df.filter(pl.col("type") == deptype)[self.index_col].to_list()

    @property
    def attachment_ids(self) -> list[str]:
        r"""Attachment IDs.

        Returns:
            list of attachment IDs

        """
        return self._df.filter(pl.col("type") == define.DependType.ATTACHMENT)[
            "archive"
        ].to_list()

    @property
    def files(self) -> list[str]:
        r"""All media, table, attachments.

        Returns:
            list of files

        """
        return list(self._idx.keys())

    @property
    def media(self) -> list[str]:
        r"""Media files.

        Returns:
            list of media

        """
        return self._df.filter(pl.col("type") == define.DependType.MEDIA)[
            self.index_col
        ].to_list()

    @property
    def removed_media(self) -> list[str]:
        r"""Removed media files.

        Returns:
            list of media

        """
        return self._df.filter(
            (pl.col("type") == define.DependType.MEDIA) & (self._df["removed"] == 1)
        )[self.index_col].to_list()

    @property
    def table_ids(self) -> list[str]:
        r"""Table IDs.

        Like :meth:`audb.Dependencies.tables`,
        but only returns the table ID,
        i.e. ``db.<id>.csv``.

        Returns:
            list of table IDs

        """
        return [table[3:-4] for table in self.tables]

    @property
    def tables(self) -> list[str]:
        r"""Tables files.

        Returns:
            list of tables

        """
        return self._df.filter(pl.col("type") == define.DependType.META)[
            self.index_col
        ].to_list()

    def archive(self, file: str) -> str:
        r"""Name of archive the file belongs to.

        Args:
            file: relative file path

        Returns:
            archive name

        Example: 'archive-0'
        """
        # return self._df.filter(pl.col("file") == file).item(0, "archive")
        # pandas implementation
        # return self._df.archive[file]
        return self._df.row(self._idx[file], named=True)["archive"]

    def bit_depth(self, file: str) -> int:
        r"""Bit depth of media file.

        Args:
            file: relative file path

        Returns:
            bit depth

        """
        # works for both pandas and polars
        return self._column_loc("bit_depth", file, int)

    def channels(self, file: str) -> int:
        r"""Number of channels of media file.

        Args:
            file: relative file path

        Returns:
            number of channels

        """
        return self._column_loc("channels", file, int)

    def checksum(self, file: str) -> str:
        r"""Checksum of file.

        Args:
            file: relative file path

        Returns:
            checksum of file

        """
        return self._column_loc("checksum", file)

    def duration(self, file: str) -> float:
        r"""Duration of file.

        Args:
            file: relative file path

        Returns:
            duration in seconds

        """
        return self._column_loc("duration", file, float)

    def format(self, file: str) -> str:
        r"""Format of file.

        Args:
            file: relative file path

        Returns:
            file format (always lower case)

        """
        return self._column_loc("format", file)

    def _update_idx(self):
        """Update index as polars does not have any."""
        self._idx = dict(
            zip(self._df[self.index_col], range(len(self._df[self.index_col])))
        )

    def load(self, path: str):
        r"""Read dependencies from file.

        Clears existing dependencies.

        Args:
            path: path to file.
                File extension can be ``csv``
                ``pkl``,
                or ``parquet``

        Raises:
            ValueError: if file extension is not one of
                ``csv``, ``pkl``, ``parquet``
            FileNotFoundError: if ``path`` does not exists

        """
        path = audeer.path(path)
        extension = audeer.file_extension(path)
        if extension not in ["csv", "pkl", "parquet"]:
            raise ValueError(
                f"File extension of 'path' has to be 'csv', 'pkl', or 'parquet' "
                f"not '{extension}'"
            )
        if not os.path.exists(path):
            raise FileNotFoundError(
                errno.ENOENT,
                os.strerror(errno.ENOENT),
                path,
            )
        if extension == "pkl":
            self._df = pd.read_pickle(path)
            # Correct dtype of index
            # to make backward compatiple
            # with old pickle files in cache
            # that might use `string` as dtype
            if self._df.index.dtype != define.DEPEND_INDEX_DTYPE:
                self._df.index = self._df.index.astype(define.DEPEND_INDEX_DTYPE)

        elif extension == "csv":
            table = csv.read_csv(
                path,
                read_options=csv.ReadOptions(
                    column_names=self._schema.names,
                    skip_rows=1,
                ),
                convert_options=csv.ConvertOptions(column_types=self._schema),
            )
            self._df = self._table_to_dataframe(table)

        elif extension == "parquet":
            table = parquet.read_table(path)
            # path
            # self._df = self._table_to_dataframe(table)
            # self._df = pl.read_parquet(path)
            # there is no index!
            # self._df = pl.read_parquet(path).rename(
            #     {"__index_level_0__": self.index_col}
            # )
            # new polars version reads index (name file) directly into variable
            self._df = pl.read_parquet(path)
            # need to maintain an index
            self._update_idx()

    def removed(self, file: str) -> bool:
        r"""Check if file is marked as removed.

        Args:
            file: relative file path

        Returns:
            ``True`` if file was removed

        """
        return self._column_loc("removed", file, bool)

    def sampling_rate(self, file: str) -> int:
        r"""Sampling rate of media file.

        Args:
            file: relative file path

        Returns:
            sampling rate in Hz

        """
        return self._column_loc("sampling_rate", file, int)

    def save(self, path: str):
        r"""Write dependencies to file.

        Args:
            path: path to file.
                File extension can be ``csv``, ``pkl``, or ``parquet``

        """
        path = audeer.path(path)
        if path.endswith("csv"):
            table = self._dataframe_to_table(self._df)
            csv.write_csv(
                table,
                path,
                write_options=csv.WriteOptions(quoting_style="none"),
            )
        elif path.endswith("pkl"):
            self._df.to_pickle(
                path,
                protocol=4,  # supported by Python >= 3.4
            )
        elif path.endswith("parquet"):
            table = self._dataframe_to_table(self._df, file_column=True)
            parquet.write_table(table, path)

    def type(self, file: str) -> int:
        r"""Type of file.

        Args:
            file: relative file path

        Returns:
            type

        """
        return self._column_loc("type", file, int)

    def version(self, file: str) -> str:
        r"""Version of file.

        Args:
            file: relative file path

        Returns:
            version string

        """
        return self._column_loc("version", file)

    def _add_attachment(
        self,
        file: str,
        version: str,
        archive: str,
        checksum: str,
    ):
        r"""Add or update attachment.

        Args:
            file: relative path of attachment
            version: version string
            archive: archive name without extension
            checksum: checksum of file

        Note that this is considered unstable, see documentation
        we use  pl.DataFrame.update

        """
        format = audeer.file_extension(file).lower()
        entry = [
            file,
            archive,  # archive
            0,  # bit_depth
            0,  # channels
            checksum,  # checksum
            0.0,  # duration
            format,  # format
            0,  # removed
            0,  # sampling_rate
            define.DependType.ATTACHMENT,  # type
            version,  # version
        ]

        self._df = self._df.update(
            pl.from_records([entry], schema=self._df.schema, orient="row"),
            how="full",
            left_on=["file"],
            right_on=["file"],
        )

        # update index only if a new attachmenbt is coming in
        if file not in self._idx:
            self._idx[file] = len(self._idx)

    def _add_media(
        self,
        values: Sequence[
            tuple[
                str,  # file
                str,  # archive
                int,  # bit_depth
                int,  # channels
                str,  # checksum
                float,  # duration
                str,  # format
                int,  # removed
                float,  # sampling_rate
                int,  # type
                str,  # version
            ]
        ],
    ):
        r"""Add media files.

        Args:
            values: list of tuples,
                where each tuple holds the values of a new media entry

        """
        self._df = pl.concat(
            [self._df, pl.from_records(values, schema=self._df.schema, orient="row")],
            how="vertical_relaxed",
        )

    def _add_meta(
        self,
        file: str,
        version: str,
        archive: str,
        checksum: str,
    ):
        r"""Add or update table file.

        Args:
            file: relative file path
            archive: archive name without extension
            checksum: checksum of file
            version: version string

        """
        format = audeer.file_extension(file).lower()

        entry = [
            file,
            archive,  # archive
            0,  # bit_depth
            0,  # channels
            checksum,  # checksum
            0.0,  # duration
            format,  # format
            0,  # removed
            0,  # sampling_rate
            define.DependType.META,  # type
            version,  # version
        ]

        self._df = self._df.update(
            pl.from_records([entry], schema=self._df.schema, orient="row"),
            how="full",
            left_on=["file"],
            right_on=["file"],
        )

        # update index only if a new attachmenbt is coming in
        if file not in self._idx:
            self._idx[file] = len(self._idx)

    def _column_loc(
        self,
        column: str,
        file: str,
        dtype: Callable = None,
    ) -> object:
        r"""Column content for selected files and column."""
        value = self._df.row(self._idx[file], named=True)[column]
        if dtype is not None:
            value = dtype(value)
        return value

    def _dataframe_to_table(
        self,
        df: pd.DataFrame,
        *,
        file_column: bool = False,
    ) -> pa.Table:
        r"""Convert pandas dataframe to pyarrow table.

        Args:
            df: dependency table as pandas dataframe
            file_column: if ``False``
                the ``"file"`` column
                is renamed to ``""``

        Returns:
            dependency table as pyarrow table

        """
        table = pa.Table.from_pandas(
            df.reset_index().rename(columns={"index": "file"}),
            preserve_index=False,
            schema=self._schema,
        )
        if not file_column:
            columns = table.column_names
            columns = ["" if c == "file" else c for c in columns]
            table = table.rename_columns(columns)
        return table

    def _drop(self, files: Sequence[str]):
        r"""Drop files from table.

        Args:
            files: relative file paths

        """
        # self._df.drop is slow,
        # see https://stackoverflow.com/a/53394627.
        # The solution presented in https://stackoverflow.com/a/53395360
        # self._df = self._df.loc[self._df.index.drop(files)]
        # which is claimed to be faster,
        # isn't.
        # pandas:
        # self._df = self._df[~self._df.index.isin(files)]

        # polars
        self._df = self._df.filter(~self._df[self.index_col].is_in(files))

    def _remove(self, file: str):
        r"""Mark file as removed.

        Args:
            file: relative file path

        """
        self._df.with_columns(
            pl.when(pl.col("file") == file)
            .then(1)
            .otherwise(pl.col("removed"))
            .alias("removed")
        )

    def _table_to_dataframe(self, table: pa.Table) -> pd.DataFrame:
        r"""Convert pyarrow table to pandas dataframe.

        Args:
            table: dependency table as pyarrow table

        Returns:
            dependency table as pandas dataframe

        """
        df = table.to_pandas(
            deduplicate_objects=False,
            # Convert to pyarrow dtypes,
            # but ensure we use pd.StringDtype("pyarrow")
            # and not pd.ArrowDtype(pa.string())
            # see https://pandas.pydata.org/docs/user_guide/pyarrow.html
            types_mapper={
                pa.string(): pd.StringDtype("pyarrow"),
                pa.int32(): pd.ArrowDtype(pa.int32()),
                pa.float64(): pd.ArrowDtype(pa.float64()),
            }.get,  # we have to provide a callable, not a dict
        )
        df.set_index("file", inplace=True)
        df.index.name = None
        df.index = df.index.astype(define.DEPEND_INDEX_DTYPE)
        return df

    def _update_media(
        self,
        values: Sequence[
            tuple[
                str,  # file
                str,  # archive
                int,  # bit_depth
                int,  # channels
                str,  # checksum
                float,  # duration
                str,  # format
                int,  # removed
                float,  # sampling_rate
                int,  # type
                str,  # version
            ]
        ],
    ):
        r"""Update media files.

        Args:
            values: list of tuples,
                where each tuple holds the new values for a media entry

        """
        # Slow implementations:
        #
        # newer concat
        # self._df = pl.concat(
        #     [self._df, pl.from_records(values, schema=self._df.schema)],
        #     how="vertical_relaxed",
        # )

        # predicate replacement example
        # self._df.with_columns(
        #     pl.when(pl.col("file") == file)
        #     .then(1)
        #     .otherwise(pl.col("removed")).alias("removed")
        # )

        self._df = self._df.update(
            pl.from_records(values, schema=self._df.schema, orient="row")
        )
        # no need to update self._idx as this is updating existing media

    def _update_media_version(
        self,
        files: Sequence[str],
        version: str,
    ):
        r"""Update version of media files.

        Args:
            files: relative file paths
            version: version string

        """
        # NOTE: pandas way of counting index
        field = define.DEPEND_FIELD_NAMES[define.DependField.VERSION]
        # using new data allocation!
        self._df = pl.concat(
            [
                self._df.filter(~pl.col(self.index_col).is_in(files)),
                self._df.filter(pl.col(self.index_col).is_in(files)).with_columns(
                    pl.col(field).str.replace(".*", version)
                ),
            ],
            how="vertical",
        )
        # no need to update self._idx as only updating


def error_message_missing_object(
    object_type: str,
    missing_object_id: str | Sequence,
    database_name: str = None,
    database_version: str = None,
) -> str:
    r"""Error message for missing objects.

    Args:
        object_type: object that is supposed to contain ``missing_object_id``,
            should be
            ``'media'``,
            ``'table'``
            or ``'attachment'``
        missing_object_id: ID of missing object
        database_name: name of affected database
        database_version: name of affected database

    Returns:
        error message

    """
    if object_type == "media":
        object_name = f"{object_type} file"
    else:
        object_name = object_type

    if isinstance(missing_object_id, str):
        msg = f"Could not find a {object_name} " f"matching '{missing_object_id}'"
    else:
        msg = f"Could not find the {object_name} " f"'{missing_object_id[0]}'"
    if database_name is not None and database_version is not None:
        msg += f" in {database_name} v{database_version}"
    return msg


def filter_deps(
    requested_deps: str | Sequence[str] | None,
    available_deps: Sequence[str],
    deps_type: str,
    database_name: str = None,
    database_version: str = None,
) -> Sequence[str]:
    r"""Filter dependency files by requested files.

    Args:
        requested_deps: include only media, tables
            matching the regular expression
            or provided in the list
        available_deps: sequence of available media files or tables
        deps_type: ``'attachment'``, ``'media'`` or ``'table'``
        database_name: name of affected database
        database_version: name of affected database

    Returns:
        list of attachments, media or tables
            inside the dependency object
            matching ``requested_deps``

    """
    if requested_deps is None:
        return available_deps
    elif len(requested_deps) == 0:
        return []

    if isinstance(requested_deps, str):
        request = requested_deps
        pattern = re.compile(request)
        requested_deps = []
        for dep in available_deps:
            if pattern.search(dep):
                requested_deps.append(dep)
        if len(requested_deps) == 0:
            msg = error_message_missing_object(
                deps_type,
                request,
                database_name,
                database_version,
            )
            raise ValueError(msg)
    else:
        for dep in requested_deps:
            if dep not in available_deps:
                msg = error_message_missing_object(
                    deps_type,
                    [dep],
                    database_name,
                    database_version,
                )
                raise ValueError(msg)

    return requested_deps


def download_dependencies(
    backend_interface: type[audbackend.interface.Base],
    name: str,
    version: str,
    verbose: bool,
) -> Dependencies:
    r"""Load dependency file from backend.

    Download dependency file
    for requested database
    to a temporary folder,
    and return an dependency object
    loaded from that file.

    Args:
        backend_interface: backend interface
        name: database name
        version: database version
        verbose: if ``True`` a message is shown during download

    Returns:
        dependency object

    """
    with tempfile.TemporaryDirectory() as tmp_root:
        # Load `db.parquet` file,
        # or if non-existent `db.zip`
        # from backend
        remote_deps_file = backend_interface.join("/", name, define.DEPENDENCIES_FILE)
        if backend_interface.exists(remote_deps_file, version):
            local_deps_file = os.path.join(tmp_root, define.DEPENDENCIES_FILE)
            backend_interface.get_file(
                remote_deps_file,
                local_deps_file,
                version,
                verbose=verbose,
            )
        else:
            remote_deps_file = backend_interface.join("/", name, define.DB + ".zip")
            local_deps_file = os.path.join(
                tmp_root,
                define.LEGACY_DEPENDENCIES_FILE,
            )
            backend_interface.get_archive(
                remote_deps_file,
                tmp_root,
                version,
                verbose=verbose,
            )
        # Create deps object from downloaded file
        deps = Dependencies()
        deps.load(local_deps_file)
    return deps


def upload_dependencies(
    backend_interface: type[audbackend.interface.Base],
    deps: Dependencies,
    db_root: str,
    name: str,
    version: str,
):
    r"""Upload dependency file to backend.

    Store a dependency file
    in the local database root folder,
    and upload it to the backend.

    Args:
        backend_interface: backend interface
        deps: dependency object
        db_root: database root folder
        name: database name
        version: database version

    """
    local_deps_file = os.path.join(db_root, define.DEPENDENCIES_FILE)
    remote_deps_file = backend_interface.join("/", name, define.DEPENDENCIES_FILE)
    deps.save(local_deps_file)
    backend_interface.put_file(local_deps_file, remote_deps_file, version)



import os
import re

import pandas as pd
import pytest

import audeer

import audb


CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
ROWS = [
    {
        "file": "db.files.csv",
        "archive": "archive1",
        "bit_depth": 0,
        "channels": 0,
        "checksum": "7c1f6b568f7221ab968a705fd5e7477b",
        "duration": 0.0,
        "format": "csv",
        "removed": 0,
        "sampling_rate": 0,
        "type": 0,
        "version": "1.0.0",
    },
    {
        "file": "db.speaker.parquet",
        "archive": "",
        "bit_depth": 0,
        "channels": 0,
        "checksum": "3hf774jkf7hfjjg775678djjd5e7dfh3",
        "duration": 0.0,
        "format": "parquet",
        "removed": 0,
        "sampling_rate": 0,
        "type": 0,
        "version": "1.0.0",
    },
    {
        "file": "file.wav",
        "archive": "archive2",
        "bit_depth": 16,
        "channels": 2,
        "checksum": "917338b854ad9c72f76bc9a68818dcd8",
        "duration": 1.23,
        "format": "wav",
        "removed": 0,
        "sampling_rate": 16000,
        "type": 1,
        "version": "1.0.0",
    },
]


def get_entries(column):
    return [row[column] for row in ROWS]


def test_get_entries():
    assert get_entries("archive") == ["archive1", "", "archive2"]


@pytest.fixture(
    scope="function",
)
def deps():
    deps = audb.Dependencies()
    df = pd.DataFrame.from_records(ROWS)
    df.set_index("file", inplace=True)
    # Ensure correct dtype
    df.index = df.index.astype(audb.core.define.DEPENDENCY_INDEX_DTYPE)
    df.index.name = None
    df = df.astype(audb.core.define.DEPENDENCY_TABLE)
    deps._df = df
    return deps


def test_instantiation():
    r"""Test instantiation of audb.Dependencies.

    During instantiation of ``audb.Dependencies``
    an empty dataframe is created under ``self._df``,
    that stores the dependency table.
    This test ensures,
    that the dataframe
    contains the correct column names and data types,
    and the correct name and data type of its index.

    """
    deps = audb.Dependencies()
    expected_columns = [
        "archive",
        "bit_depth",
        "channels",
        "checksum",
        "duration",
        "format",
        "removed",
        "sampling_rate",
        "type",
        "version",
    ]
    expected_df = pd.DataFrame(columns=expected_columns)
    expected_df.index = expected_df.index.astype(
        audb.core.define.DEPENDENCY_INDEX_DTYPE
    )
    expected_df = expected_df.astype(audb.core.define.DEPENDENCY_TABLE)
    pd.testing.assert_frame_equal(deps._df, expected_df)
    assert list(deps._df.columns) == expected_columns
    df = deps()
    assert list(df.columns) == expected_columns


def test_call(deps):
    expected_df = pd.DataFrame.from_records(ROWS).set_index("file")
    expected_df.index = expected_df.index.astype(
        audb.core.define.DEPENDENCY_INDEX_DTYPE
    )
    expected_df.index.name = None
    expected_df = expected_df.astype(audb.core.define.DEPENDENCY_TABLE)
    df = deps()
    pd.testing.assert_frame_equal(df, expected_df)


def test_contains(deps):
    assert "db.files.csv" in deps
    assert "file.wav" in deps
    assert "db.speaker.parquet" in deps
    assert "not.csv" not in deps


def test_equals(deps):
    # empty table vs. empty table
    _deps = audb.Dependencies()
    assert _deps == _deps
    assert _deps == audb.Dependencies()
    # empty table vs. example table
    assert deps != audb.Dependencies()
    # example table vs. example table
    assert deps == deps
    _deps._df = deps._df.copy()
    assert deps == _deps
    # example table vs. different table
    _deps._df.loc["db.files.csv", "channels"] = 4
    assert deps != _deps


def test_get_item(deps):
    assert deps["db.files.csv"] == list(ROWS[0].values())[1:]
    print(deps["db.files.csv"])
    print(deps().archive)
    assert deps["db.speaker.parquet"] == list(ROWS[1].values())[1:]
    assert deps["file.wav"] == list(ROWS[2].values())[1:]
    with pytest.raises(KeyError, match="non.existing"):
        deps["non.existing"]


def test_archives(deps):
    assert deps.archives == sorted(get_entries("archive"))


def test_files(deps):
    assert deps.files == get_entries("file")


def test_media(deps):
    assert deps.media == ["file.wav"]


def test_removed_media(deps):
    assert deps.removed_media == []


def test_table_ids(deps):
    assert deps.table_ids == ["files", "speaker"]


def test_tables(deps):
    assert deps.tables == ["db.files.csv", "db.speaker.parquet"]


def test_archive(deps):
    files = get_entries("file")
    archives = get_entries("archive")
    for file, archive in zip(files, archives):
        assert deps.archive(file) == archive
        assert isinstance(deps.archive(file), str)
    with pytest.raises(KeyError, match="non.existing"):
        deps.archive("non.existing")


def test_bit_depth(deps):
    files = get_entries("file")
    bit_depths = get_entries("bit_depth")
    for file, bit_depth in zip(files, bit_depths):
        assert deps.bit_depth(file) == bit_depth
        assert isinstance(deps.bit_depth(file), int)
    with pytest.raises(KeyError, match="non.existing"):
        deps.bit_depth("non.existing")


def test_channels(deps):
    files = get_entries("file")
    channels = get_entries("channels")
    for file, channel in zip(files, channels):
        assert deps.channels(file) == channel
        assert isinstance(deps.channels(file), int)
    with pytest.raises(KeyError, match="non.existing"):
        deps.channels("non.existing")


def test_checksum(deps):
    files = get_entries("file")
    checksums = get_entries("checksum")
    for file, checksum in zip(files, checksums):
        assert deps.checksum(file) == checksum
        assert isinstance(deps.checksum(file), str)
    with pytest.raises(KeyError, match="non.existing"):
        deps.checksum("non.existing")


def test_duration(deps):
    files = get_entries("file")
    durations = get_entries("duration")
    for file, duration in zip(files, durations):
        assert deps.duration(file) == duration
        assert isinstance(deps.duration(file), float)
    with pytest.raises(KeyError, match="non.existing"):
        deps.duration("non.existing")


def test_format(deps):
    files = get_entries("file")
    formats = get_entries("format")
    for file, format in zip(files, formats):
        assert deps.format(file) == format
        assert isinstance(deps.format(file), str)
    with pytest.raises(KeyError, match="non.existing"):
        deps.format("non.existing")


def test_removed(deps):
    files = get_entries("file")
    removeds = get_entries("removed")
    for file, removed in zip(files, removeds):
        assert deps.removed(file) == removed
        assert isinstance(deps.removed(file), bool)
    with pytest.raises(KeyError, match="non.existing"):
        deps.removed("non.existing")


@pytest.mark.parametrize("file", ["deps.csv", "deps.pkl", "deps.parquet"])
def test_load_save(tmpdir, deps, file):
    """Test consistency of dependency table after save/load cycle.

    Dependency values and data types
    should remain identical
    when first storing and then loading from a file.
    This should hold for all possible file formats.

    """
    deps_file = audeer.path(tmpdir, file)
    deps.save(deps_file)
    deps2 = audb.Dependencies()
    deps2.load(deps_file)
    pd.testing.assert_frame_equal(deps(), deps2())
    assert list(deps2._df.dtypes) == list(audb.core.define.DEPENDENCY_TABLE.values())


def test_load_save_backward_compatibility(tmpdir, deps):
    """Test backward compatibility with old pickle cache files.

    As the dtype of the index has changed,
    we need to make sure this is corrected
    when loading old cache files.

    Old behaviour (audb<1.7):

    archive          string[python]
    bit_depth                 int32
    channels                  int32
    checksum         string[python]
    duration                float64
    format           string[python]
    removed                   int32
    sampling_rate             int32
    type                      int32
    version          string[python]

    New behaviour (audb>=1.7):

    archive          string[pyarrow]
    bit_depth         int32[pyarrow]
    channels          int32[pyarrow]
    checksum         string[pyarrow]
    duration         double[pyarrow]
    format           string[pyarrow]
    removed           int32[pyarrow]
    sampling_rate     int32[pyarrow]
    type              int32[pyarrow]
    version          string[pyarrow]

    """
    deps_file = audeer.path(tmpdir, "deps.pkl")

    deps_old = audb.Dependencies()
    deps_old._df = deps._df.copy()

    # Change dtype of index from object to string
    # to mimic previous behavior
    deps_old._df.index = deps_old._df.index.astype("string")
    # Change dtype of columns
    # to mimic previous behavior
    deps_old._df = deps_old._df.astype(
        {
            "archive": "string",
            "bit_depth": "int32",
            "channels": "int32",
            "checksum": "string",
            "duration": "float64",
            "format": "string",
            "removed": "int32",
            "sampling_rate": "int32",
            "type": "int32",
            "version": "string",
        }
    )
    deps_old.save(deps_file)

    # Check that we get the correct dtypes,
    # when loading from cache
    deps2 = audb.Dependencies()
    deps2.load(deps_file)
    assert deps2._df.index.dtype == audb.core.define.DEPENDENCY_INDEX_DTYPE
    pd.testing.assert_frame_equal(deps._df, deps2._df)
    assert deps == deps2


@pytest.mark.parametrize("pandas_version", ["2.0.3", "2.1.4", "2.2.2"])
def test_load_save_pandas_compatibility(pandas_version):
    """Test pandas backward compatibility of pickle cache files.

    Dataframes using pyarrow dtypes,
    and stored as pickle files
    might fail to load
    if the used ``pandas`` version
    does not match.

    Test which ``pandas`` versions raise errors
    when loading deps from pickle files.
    We have to except those errors inside
    ``audb.dependencies()``.

    See Also:
    https://github.com/audeering/audb/issues/418

    Args:
        pandas_version: the version of ``pandas``
            used to store the dependency table in cache

    """
    deps_file = audeer.path(
        CURRENT_DIR,
        "assests",
        "dependency-table-pandas",
        f"emodb-pandas-{pandas_version}.pkl",
    )
    deps = audb.Dependencies()

    # Dependency table cached with pandas==2.0.3.
    # Loading with pandas>=2.1.0 leads to a ModuleNotFoundError
    if pd.__version__ >= "2.1.0" and pandas_version == "2.0.3":
        error_msg = "No module named 'pandas.core.arrays.arrow.dtype'"
        with pytest.raises(ModuleNotFoundError, match=error_msg):
            deps.load(deps_file)

    # Dependency table cached with pandas>=2.1.4.
    # Loading with pandas==2.0.3 leads to a KeyError
    elif pd.__version__ == "2.0.3" and pandas_version >= "2.1.4":
        error_msg = "'_data'"
        with pytest.raises(KeyError, match=error_msg):
            deps.load(deps_file)

    else:
        deps.load(deps_file)
        assert deps._df.index.dtype == audb.core.define.DEPENDENCY_INDEX_DTYPE


def test_load_save_errors(deps):
    """Test possible errors when loading/saving."""
    # Wrong file extension
    with pytest.raises(ValueError, match=r".*'txt'.*"):
        deps.load("deps.txt")
    # File missing
    with pytest.raises(FileNotFoundError):
        deps.load("deps.csv")


def test_sampling_rate(deps):
    files = get_entries("file")
    sampling_rates = get_entries("sampling_rate")
    for file, sampling_rate in zip(files, sampling_rates):
        assert deps.sampling_rate(file) == sampling_rate
        assert isinstance(deps.sampling_rate(file), int)
    with pytest.raises(KeyError, match="non.existing"):
        deps.sampling_rate("non.existing")


def test_type(deps):
    files = get_entries("file")
    types = get_entries("type")
    for file, type in zip(files, types):
        assert deps.type(file) == type
        assert isinstance(deps.type(file), int)
    with pytest.raises(KeyError, match="non.existing"):
        deps.type("non.existing")


def test_version(deps):
    files = get_entries("file")
    versions = get_entries("version")
    for file, version in zip(files, versions):
        assert deps.version(file) == version
        assert isinstance(deps.version(file), str)
    with pytest.raises(KeyError, match="non.existing"):
        deps.version("non.existing")


def test_len(deps):
    assert len(deps) == len(ROWS)


def test_str(deps):
    r"""Test string representation of dependency table."""
    # Check only parts of the returned string,
    # as the representation might vary,
    # see https://github.com/audeering/audb/issues/422
    expected_str = re.compile(
        "                     archive  bit_depth  channels .+? type version\n"
        "db.files.csv        archive1          0         0 .+? 0   1.0.0\n"
        "db.speaker.parquet                    0         0 .+? 0   1.0.0\n"
        "file.wav            archive2         16         2 .+? 1   1.0.0.*?"
    )
    print(str(deps))
    assert expected_str.match(str(deps))
    assert expected_str.match(deps._df.to_string())


# === Test hidden methods ===
@pytest.mark.parametrize(
    "file, version, archive, checksum",
    [
        ("attachment.txt", "1.0.0", "andhfner", "asndmsknfporkrgfk3l"),
    ],
)
def test_add_attachment(deps, file, version, archive, checksum):
    deps._add_attachment(file, version, archive, checksum)
    assert len(deps) == len(ROWS) + 1
    assert deps.version(file) == version
    assert deps.archive(file) == archive
    assert deps.checksum(file) == checksum


@pytest.mark.parametrize(
    "values",
    [
        [
            (
                "file1.wav",
                "archive1",
                16,
                1,
                "jsdfjioergjiergnmo",
                2.3,
                "wav",
                0,
                16000,
                audb.core.define.DEPENDENCY_TYPE["media"],
                "1.1.0",
            ),
            (
                "file2.wav",
                "archive2",
                24,
                1,
                "masdfmiosedascrf34",
                5.6,
                "wav",
                0,
                44100,
                audb.core.define.DEPENDENCY_TYPE["media"],
                "1.2.0",
            ),
        ],
    ],
)
def test_add_media(deps, values):
    deps._add_media(values)
    assert len(deps) == len(ROWS) + 2  # as we added already attachment before
    for (
        file,
        archive,
        bit_depth,
        channels,
        checksum,
        duration,
        format,
        removed,
        sampling_rate,
        type,
        version,
    ) in values:
        assert deps.archive(file) == archive
        assert deps.bit_depth(file) == bit_depth
        assert deps.channels(file) == channels
        assert deps.checksum(file) == checksum
        assert deps.duration(file) == duration
        assert deps.format(file) == format
        assert deps.removed(file) == removed
        assert deps.sampling_rate(file) == sampling_rate
        assert deps.type(file) == type
        assert deps.version(file) == version


@pytest.mark.parametrize(
    "file, version, checksum, expected_archive",
    [
        ("db.table1.csv", "2.1.0", "asddfnfpork45rgfl", "table1"),
        ("db.table1.parquet", "2.1.0", "asddfnfpork45rgfl", ""),
    ],
)
def test_add_meta(deps, file, version, checksum, expected_archive):
    deps._add_meta(file, version, checksum)
    assert len(deps) == len(ROWS) + 1
    assert deps.version(file) == version
    assert deps.archive(file) == expected_archive
    assert deps.checksum(file) == checksum


@pytest.mark.parametrize(
    "files, expected_length",
    [
        (["file.wav"], 2),
        (["db.files.csv"], 2),
        (["file.wav", "db.files.csv"], 1),
        (["file.wav", "db.files.csv", "db.speaker.parquet"], 0),
    ],
)
def test_drop(deps, files, expected_length):
    deps._drop(files)
    assert len(deps) == expected_length
    for file in files:
        assert file not in deps


@pytest.mark.parametrize(
    "file",
    [
        ("file.wav"),
        ("db.files.csv"),
    ],
)
def test_remove(deps, file):
    assert not deps.removed(file)
    deps._remove(file)
    assert len(deps) == len(ROWS)
    assert deps.removed(file)


@pytest.mark.parametrize(
    "values",
    [
        [
            (
                "file.wav",
                "archive1",
                16,
                1,
                "jsdfjioergjiergnmo",
                2.3,
                "wav",
                0,
                16000,
                audb.core.define.DEPENDENCY_TYPE["media"],
                "1.1.0",
            ),
        ],
        pytest.param(
            [
                (
                    "non-existent.wav",
                    "archive1",
                    16,
                    1,
                    "jsdfjioergjiergnmo",
                    2.3,
                    "wav",
                    0,
                    16000,
                    audb.core.define.DEPENDENCY_TYPE["media"],
                    "1.1.0",
                ),
            ],
            marks=pytest.mark.xfail(raises=KeyError),
        ),
    ],
)
def test_update_media(deps, values):
    deps._update_media(values)
    assert len(deps) == len(ROWS)
    for (
        file,
        archive,
        bit_depth,
        channels,
        checksum,
        duration,
        format,
        removed,
        sampling_rate,
        type,
        version,
    ) in values:
        assert deps.archive(file) == archive
        assert deps.bit_depth(file) == bit_depth
        assert deps.channels(file) == channels
        assert deps.checksum(file) == checksum
        assert deps.duration(file) == duration
        assert deps.format(file) == format
        assert deps.removed(file) == removed
        assert deps.sampling_rate(file) == sampling_rate
        assert deps.type(file) == type
        assert deps.version(file) == version


@pytest.mark.parametrize(
    "files, version",
    [
        (["file.wav"], "3.0.0"),
        (["file.wav", "db.files.csv"], "4.0.0"),
        pytest.param(
            ["non-existent.wav"],
            "3.0.0",
            marks=pytest.mark.xfail(raises=KeyError),
        ),
    ],
)
def test_update_media_version(deps, files, version):
    deps._update_media_version(files, version)
    assert len(deps) == len(ROWS)
    for file in files:
        assert deps.version(file) == version



import glob
import os
import platform
import sys

import pytest

import audeer

import audb


PUBLIC_HOST = "s3.dualstack.eu-north-1.amazonaws.com"
PUBLIC_BACKEND = "s3"


if platform.system() == "Darwin":
    # Avoid multi-threading on MacOS runner,
    # as it might fail from time to time
    # (memory issue?), see
    # https://github.com/audeering/audresample/issues/57
    pytest.NUM_WORKERS = 1
else:
    pytest.NUM_WORKERS = 5


@pytest.fixture(scope="package", autouse=True)
def cleanup_coverage_files():
    path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        ".coverage.*",
    )
    for file in glob.glob(path):
        os.remove(file)


# ===== CACHE =====
@pytest.fixture(scope="function", autouse=True)
def cache(tmpdir_factory):
    r"""Temp folder as cache.

    Provide a different temporary folder
    as audb cache root
    in each test.

    """
    cache = tmpdir_factory.mktemp("cache")
    current_cache = audb.config.CACHE_ROOT
    audb.config.CACHE_ROOT = str(cache)

    yield cache

    audb.config.CACHE_ROOT = current_cache


@pytest.fixture(scope="module", autouse=True)
def persistent_cache(tmpdir_factory):
    r"""Temp folder as module wide cache.

    Provide a different temporary folder
    as cache across all tests
    in a test definition file (module).

    This cache will be used automatically
    in all fixtures
    that have module as scope
    and access the cache folder.

    """
    cache = tmpdir_factory.mktemp("cache")
    current_cache = audb.config.CACHE_ROOT
    audb.config.CACHE_ROOT = str(cache)

    yield cache

    audb.config.CACHE_ROOT = current_cache


@pytest.fixture(scope="package", autouse=True)
def shared_cache(tmpdir_factory):
    r"""Temp folder as shared cache.

    Provide a single temporary folder
    as audb shared cache root
    across all tests.

    """
    cache = tmpdir_factory.mktemp("shared_cache")
    audb.config.SHARED_CACHE_ROOT = str(cache)
    return cache


@pytest.fixture(scope="package", autouse=True)
def hide_default_caches():
    r"""Hide default audb cache settings during testing.

    The cache and shared cache of audb
    are handled by ``audb.config.CACHE_ROOT``,
    ``audb.config.SHARED_CACHE_ROOT``
    and can in addition be configured
    by the environment variables
    ``AUDB_CACHE_ROOT``
    and ``AUDB_SHARED_CACHE_ROOT``.

    To ensure those will not interfere with the tests
    we hide them when executing the tests.

    """
    audb.config.CACHE_ROOT = None
    # Don't clean audb.config.SHARED_CACHE_ROOT
    # as it is set by shared_cache anyway

    env_cache = os.environ.get("AUDB_CACHE_ROOT", None)
    env_shared_cache = os.environ.get("AUDB_SHARED_CACHE_ROOT", None)
    if env_cache is not None:
        del os.environ["AUDB_CACHE_ROOT"]
    if env_shared_cache is not None:
        del os.environ["AUDB_SHARED_CACHE_ROOT"]

    yield

    if env_cache is not None:
        os.environ["AUDB_CACHE_ROOT"] = env_cache
    if env_shared_cache is not None:
        os.environ["AUDB_SHARED_CACHE_ROOT"] = env_shared_cache


# ===== REPOSITORIES =====
@pytest.fixture(scope="function", autouse=False)
def repository(tmpdir_factory):
    r"""Temp folder as repository.

    Provide a different temporary folder
    as repository in each test.
    This repository will be the only one visible
    inside the test.

    """
    host = tmpdir_factory.mktemp("host")
    name = "data-unittests-local"
    repository = audb.Repository(
        name=name,
        host=host,
        backend="file-system",
    )
    audeer.mkdir(host, name)
    current_repositories = audb.config.REPOSITORIES
    audb.config.REPOSITORIES = [repository]

    yield repository

    audb.config.REPOSITORIES = current_repositories


@pytest.fixture(scope="module", autouse=False)
def persistent_repository(tmpdir_factory):
    r"""Temp folder as module wide repository.

    Provide a different temporary folder
    as repository across all tests
    in a test definition file (module).
    This repository will be the only one visible
    inside each test/fixture
    that uses it as argument.

    This fixture is useful to first publish
    one or more databases
    at the beginning of a test module
    (e.g. inside a fixture)
    and access those database(s)
    in different tests
    in the same module.

    """
    host = tmpdir_factory.mktemp("host")
    name = "data-unittests-local"
    repository = audb.Repository(
        name=name,
        host=host,
        backend="file-system",
    )
    audeer.mkdir(host, name)
    current_repositories = audb.config.REPOSITORIES
    audb.config.REPOSITORIES = [repository]

    yield repository

    audb.config.REPOSITORIES = current_repositories


@pytest.fixture(scope="module", autouse=False)
def private_and_public_repository():
    r"""Private and public repositories.

    Configure the following repositories:

    * data-private: repo on public Artifactory without access
    * audb-private: repo on public S3 without access
    * data-public: repo on public Artifactory with anonymous access
    * audb-public: repo on public S3 with anonymous access
    * data-public2: repo on public Artifactory with anonymous access

    Note, that the order of the repos is important.
    audb will visit the repos in the given order
    until it finds the requested database.

    """
    current_repositories = audb.config.REPOSITORIES
    public_artifactory_host = "https://audeering.jfrog.io/artifactory"
    public_s3_host = "s3.dualstack.eu-north-1.amazonaws.com"
    audb.config.REPOSITORIES = [
        audb.Repository("audb-private", public_s3_host, "s3"),
        audb.Repository("audb-public", public_s3_host, "s3"),
    ]
    if sys.version_info < (3, 12):
        audb.config.REPOSITORIES += [
            audb.Repository("data-private", public_artifactory_host, "artifactory"),
            audb.Repository("data-public", public_artifactory_host, "artifactory"),
            audb.Repository("data-public2", public_artifactory_host, "artifactory"),
        ]

    yield repository

    audb.config.REPOSITORIES = current_repositories


@pytest.fixture(scope="module", autouse=False)
def non_existing_repository():
    r"""Non-existing repository.

    Configure the following repositories:
    * non-existing: non-exsiting repo on public Artifactory
    * audb-public: repo on public Artifactory with anonymous access

    Note, that the order of the repos is important.
    audb will visit the repos in the given order
    until it finds the requested database.

    """
    current_repositories = audb.config.REPOSITORIES
    audb.config.REPOSITORIES = [
        audb.Repository("non-existing", PUBLIC_HOST, PUBLIC_BACKEND),
        audb.Repository("audb-public", PUBLIC_HOST, PUBLIC_BACKEND),
    ]

    yield repository

    audb.config.REPOSITORIES = current_repositories


@pytest.fixture(scope="package", autouse=True)
def hide_default_repositories():
    r"""Hide default audb repositories during testing."""
    audb.config.REPOSITORIES = []


# ===== STORAGE FORMAT =====
@pytest.fixture(scope="module", autouse=False)
def storage_format():
    """Storage format of tables."""
    yield "csv"
